"""Fitting the asymptotic relation to an SNR spectrum

This module fits the asymptotic relation to the p-modes in a frequency range
around nu_max, the central frequency of the seismic mode envelope,
in a solar-like oscillator. Only l=0 and l=2 are fit, l=1 modes are ignored.

"""

import numpy as np
import pbjam as pb
import pandas as pd
import scipy.stats as scist
from .plotting import plotting
from .jar import to_log10, normal
from collections import OrderedDict

class asymp_spec_model():
    """Class for spectrum model using asymptotic relation.

    Parameters
    ---------_
    f : float, ndarray
        Array of frequency bins of the spectrum (muHz). Truncated to the range
        around numax.
    norders : int
        Number of radial order to fit

    Attributes
    ----------
    f : float, ndarray
        Array of frequency bins of the spectrum (muHz). Truncated to the range
        around numax.

    norders : int
        Number of radial order to fit
        
    """

    def __init__(self, f, norders):
        self.f = np.array([f]).flatten()
        self.norders = int(norders)
    

    def _get_nmax(self, dnu, numax, eps):
        """Compute radial order at numax.
    
        Compute the radial order at numax, which in this implimentation of the
        asymptotic relation is not necessarily integer.
    
        Parameters
        ----------
        numax : float
            Frequency of maximum power of the p-mode envelope (muHz).
        dnu : float
            Large separation of l=0 modes (muHz).
        eps : float
            Epsilon phase term in asymptotic relation (muHz).
    
        Returns
        -------
            nmax : float
                non-integer radial order of maximum power of the p-mode envelope
                
        """
    
        return numax / dnu - eps

    def _get_enns(self, nmax, norders):
        """Compute radial order numbers.
    
        Get the enns that will be included in the asymptotic relation fit. 
        These are all integer.
    
        Parameters
        ----------
        nmax : float
            Frequency of maximum power of the p-mode envelope
        norders : int
            Total number of radial orders to consider
    
        Returns
        -------
        enns : ndarray
                Numpy array of norders radial orders (integers) around numax (nmax).
                
        """
    
        below = np.floor(nmax - np.floor(norders/2)).astype(int)
        above = np.floor(nmax + np.ceil(norders/2)).astype(int)
    
        # Handling of single input (during fitting), or array input when evaluating
        # the fit result
        if type(below) != np.ndarray:
            return np.arange(below, above)
        else:
            out = np.concatenate([np.arange(x, y) for x, y in zip(below, above)])
            return out.reshape(-1, norders)

    def _asymptotic_relation(self, numax, dnu, eps, alpha, norders):
        """ Compute the l=0 mode frequencies from the asymptotic relation for
        p-modes
    
        Parameters
        ----------
        numax : float
            Frequency of maximum power of the p-mode envelope (muHz).
        dnu : float
            Large separation of l=0 modes (muHz).
        eps : float
            Epsilon phase term in asymptotic relation (unitless).
        alpha : float
            Curvature factor of l=0 ridge (second order term, unitless).
        norders : int
            Number of desired radial orders to calculate frequncies for, centered
            around numax.
    
        Returns
        -------
        nu0s : ndarray
            Array of l=0 mode frequencies from the asymptotic relation (muHz).
    
        """
        
        nmax = self._get_nmax(dnu, numax, eps)
        enns = self._get_enns(nmax, norders)
        return (enns.T + eps + alpha/2*(enns.T - nmax)**2) * dnu


    def _P_envelope(self, nu, hmax, numax, width):
        """ Power of the seismic p-mode envelope
    
        Computes the power at frequency nu in the p-mode envelope from a Gaussian
        distribution. Used for computing mode heights.
    
        Parameters
        ----------
        nu : float
            Frequency (in muHz).
    
        hmax : float
            Height of p-mode envelope (in SNR).
    
        numax : float
            Frequency of maximum power of the p-mode envelope (in muHz).
    
        width : float
            Width of the p-mode envelope (in muHz).
    
        Returns
        -------
        h : float
            Power at frequency nu (in SNR)
            
        """
    
        return hmax * np.exp(- 0.5 * (nu - numax)**2 / width**2)

    def _lor(self, freq, h, w):
        """ Lorentzian to describe a mode.

        Parameters
        ----------
        freq : float
            Frequency of lorentzian (muHz).
        h : float
            Height of the lorentizan (SNR).
        w : float
            Full width of the lorentzian (log10(muHz)).

        Returns
        -------
        mode : ndarray
            The SNR as a function frequency for a lorentzian.
            
        """
       
        return h / (1.0 + 4.0/w**2*(self.f - freq)**2)


    def _pair(self, freq0, h, w, d02, hfac=0.7):
        """Define a pair as the sum of two Lorentzians.

        A pair is assumed to consist of an l=0 and an l=2 mode. The widths are
        assumed to be identical, and the height of the l=2 mode is scaled
        relative to that of the l=0 mode. The frequency of the l=2 mode is the
        l=0 frequency minus the small separation.

        Parameters
        ----------
        freq0 : float
            Frequency of the l=0 (muHz).
        h : float
            Height of the l=0 (SNR).
        w : float
            The mode width (identical for l=2 and l=0) (log10(muHz)).
        d02 : float
            The small separation (muHz).
        hfac : float, optional
            Ratio of the l=2 height to that of l=0 (unitless).

        Returns
        -------
        pair_model : array
            The SNR as a function of frequency of a mode pair.
            
        """
        
        pair_model = self._lor(freq0, h, w)
        pair_model += self._lor(freq0 - d02, h*hfac, w)
        return pair_model

    
    def model(self, dnu, numax, eps, d02, alpha, hmax, envwidth, modewidth,
              *args):
        """ Constructs a spectrum model from the asymptotic relation.

        The asymptotic relation for p-modes with angular degree, l=0, is
        defined as:

        nu_nl = (n + epsilon + alpha/2(n - nmax)**2) * log(dnu) ,

        where nmax = numax / dnu - epsilon.

        We separate the l=0 and l=2 modes by the small separation d02.

        Parameters
        ----------
        dnu : float
            Large separation log10(muHz)
        lognumax : float
            Frequency of maximum power of the p-mode envelope log10(muHz)

        eps : float
            Phase term of the asymptotic relation (unitless)
        alpha : float
            Curvature of the asymptotic relation log10(unitless)
        d02 : float
            Small separation log10(muHz)
        loghmax : float
            Gaussian height of p-mode envelope log10(SNR)
        logenvwidth : float
            Gaussian width of the p-mode envelope log10(muHz)
        logmodewidth : float
            Width of the modes (log10(muHz))
        *args : array-like
            List of additional parameters (Teff, bp_rp) that aren't actually
            used to construct the spectrum model, but just for evaluating the
            prior.

        Returns
        -------
        model : ndarray
            spectrum model around the p-mode envelope
            
        """

        f0s = self._asymptotic_relation(10**numax, 10**dnu, eps, 10**alpha, self.norders)
        Hs = self._P_envelope(f0s, 10**hmax, 10**numax, 10**envwidth)
        
        modewidth = 10**modewidth # widths are the same for all modes
        d02 = 10**d02
        
        mod = np.ones(len(self.f))
        for n in range(len(f0s)):
            mod += self._pair(f0s[n], Hs[n], modewidth, d02)
        return mod

    def __call__(self, p):
        """ Produce model of the asymptotic relation

        Parameters
        ----------
        p : list
            list of model parameters

        Returns
        -------
        model : array
            spectrum model around the p-mode envelope
            
        """

        return self.model(*p)


class asymptotic_fit(plotting, asymp_spec_model):
    """ Class for fitting a spectrum based on the asymptotic relation.

    Parameters
    ----------
    st : star class instance
    
    norders : int, optional
        Number of radial orders to fit

    Attributes
    ----------
    f : ndarray
        Numpy array of frequency bins of the spectrum (muHz).
    s : ndarray
        Numpy array of power in each frequency bin (SNR).
    sel : ndarray, bool
        Numpy array of boolean values specifying the frequency range to be
        considered in the asymptotic relation fit.
    model : asy_peakbag.model.model instance
        Function for computing a spectrum model given a set of parameters.
    bounds : ndarray
        Numpy array of upper and lower boundaries for the asymptotic relation
        fit. These limits truncate the likelihood function.
    gaussian : ndarray
        Numpy array of tuples of mean and sigma for Gaussian
        priors on each of the fit parameters (To be removed when full
        KDE is implimented).
        
    """

    def __init__(self, st, norders=None):
        self.pg = st.pg
        self.f = st.f
        self.s = st.s
        self.norders = norders
        self.obs = {'dnu': st.dnu, 'numax': st.numax, 'teff': st.teff, 'bp_rp': st.bp_rp}
        self.log_obs = {x: to_log10(*self.obs[x]) for x in self.obs.keys() if x != 'bp_rp'}
                        
        self.par_names = ['dnu', 'numax', 'eps', 'd02', 'alpha', 'env_height',
                          'env_width', 'mode_width', 'teff', 'bp_rp']
        self.start_samples = st.kde.samples       
        self.kde = st.kde.kde    
        self.start = self._get_asy_start()
        lfreq, ufreq = self._get_freq_range()
        self.sel = (lfreq < self.f) & (self.f < ufreq)
        self.model = asymp_spec_model(self.f[self.sel], self.norders)
        
        st.asy_fit = self
       
    def __call__(self):
        """ Setup, run and parse the asymptotic relation fit using EMCEE.

        Parameters
        ----------
        dnu : [real, real]
            Large frequency spacing and uncertainty
        numax : [real, real]
            Frequency of maximum power and uncertainty
        teff : [real, real]
            Stellar effective temperature and uncertainty
        bp_rp : [real, real]
            The Gaia Gbp - Grp color value and uncertainty
            (probably ~< 0.01 dex).

        Returns
        -------
        asy_result : Dict
            A dictionary of the modeID DataFrame and the summary DataFrame.
            
        """
        #self._start_init() # TODO - finish up this function

        self.fit = pb.mcmc(np.median(self.start_samples, axis=0), 
                           self.likelihood, self.prior)

        self.fit(start_samples=self.start_samples)


        self.modeID = self.get_modeIDs(self.fit, self.norders)
        self.summary  = self._get_summary_stats(self.fit)        
        self.mle_model = self.model(self.summary['mle'])        
        self.samples = self.fit.flatchain
        self.acceptance = self.fit.acceptance

        return {'modeID': self.modeID, 'summary': self.summary}


    def prior(self, p):
        """ Calculates the log prior 
        
        Evaluates the KDE for the parameters p. Additional hard/soft priors
        can be added here as needed to, e.g., apply boundaries to the fit.
        
        Hard constraints should be applied at the top so function exits early,
        if necessary.

        Parameters
        ----------
        p : array
            Array of model parameters

        Returns
        -------
        lp : real
            The log likelihood evaluated at p.

        """
        
        # d02/dnu < 0.2  (np.log10(0.2) ~ -0.7)
        if p[3] - p[0] > -0.7:
            return -np.inf
        
        lp = 0
        
        # Added linewidth constraints
        if (p[7] > self.start[7] + np.log10(1.5)):
            lp += normal(10**p[7], 10**self.start[7]*1.5, 10**self.start[7]*0.1)
        
        # Constraints from KDE
        lp += np.log(self.kde.pdf(p))
        
        return lp
        
    def likelihood(self, p):
        """ Likelihood function for set of model parameters

        Evaluates the likelihood function for a set of
        model parameters.  This includes the constraint from
        the observed variables.

        The code now includes a penalty to limit very large linewidth
        examples.  The penalty is very basic and in some cases where the
        true linewidth is much larger than the linewidth in the
        prior it will become informative.  For the most part this
        should not be a problem because if you  care about linewidths
        you should be using the outpout from the peakbag model.

        Parameters
        ----------
        p : array
            Array of model parameters

        Returns
        -------
        lnlike : float
            The log likelihood evaluated at p.

        """
        
        lnlike = 0
        
        # Constraint from input obs
        lnlike += normal(p[-2], *self.log_obs['teff'])
        lnlike += normal(p[-1], *self.obs['bp_rp'])
        
        # Constraint from the periodogram
        mod = self.model(p)
        lnlike += -np.sum(np.log(mod) + self.s[self.sel] / mod)
        return lnlike
    

    def _get_summary_stats(self, fit):
        """ Make dataframe with fit summary statistics
    
        Creates a dataframe that contains various quantities that summarize the
        fit. Note, these are predominantly derived from the marginalized posteriors.
       
        Parameters
        ----------
        fit : mcmc.mcmc class instance
            mcmc class instances used in the fit
            
        Returns
        -------
        summary : pandas.DataFrame
            Dataframe with the summary statistics.
            
        """
    
        idx = np.argmax(fit.flatlnlike)
        
        fc = fit.flatchain
       
        # Append here to add other statistics
        stats = OrderedDict({'mle'  : fc[idx,:], 
                             'mean' : np.mean(fc, axis = 0),
                             'std'  : np.std(fc, axis = 0),
                             'skew' : scist.skew(fc, axis = 0),
                             '2nd'  : np.percentile(fc,  2.27501, axis=0),
                             '16th' : np.percentile(fc, 15.86552, axis=0),
                             '50th' : np.percentile(fc, 50., axis=0),
                             '84th' : np.percentile(fc, 84.13447, axis=0),
                             '97th' : np.percentile(fc, 97.72498, axis=0),
                             'MAD'  : scist.median_absolute_deviation(fc, axis=0)})
    
        summary = pd.DataFrame(stats, index = self.par_names)
        
        return summary
       

    def get_modeIDs(self, fit, norders):
        """ Set mode ID in a dataframe

        Evaluates the asymptotic relation for each walker position from the
        MCMC fit. The median values of the resulting set of frequencies are
        then returned in a pandas.DataFrame

        Parameters
        ----------
        fit : mcmc.mcmc class instance
            mcmc class instances used in the fit
        norders : int
            Number of radial orders to output. Note that doesn't have to be
            the same as that used int he fit itself.

        Returns
        -------
        modeID : pandas.DataFrame
            Dataframe of radial order, n (best guess), angular degree, l,
            frequency and frequency error.
            
        """

        fc = fit.flatchain

        nu0_samps = self._asymptotic_relation(10**fc[:, 1], 10**fc[:, 0], fc[:, 2], 10**fc[:, 4], norders)
        nu2_samps = nu0_samps - 10**fc[:, 3]

        nus_med = np.median(np.array([nu0_samps, nu2_samps]), axis=2)
        nus_mad = scist.median_absolute_deviation(np.array([nu0_samps, nu2_samps]), axis=2)

        ells = np.array([2, 0]*norders) 

        df = pd.DataFrame({'ell': ells, 'nu_med': np.zeros(len(ells)), 'nu_mad': np.zeros(len(ells))})

        df.at[::2, 'nu_med'] = nus_med[1, :]
        df.at[1::2, 'nu_med'] = nus_med[0, :]
        
        df.at[::2, 'nu_mad'] = nus_mad[1, :]
        df.at[1::2, 'nu_mad'] = nus_mad[0, :]

        return df


    
    def _get_asy_start(self):
        """ Get start averages for sampling
        """
 
        mu = np.median(self.start_samples, axis=0)
        start = [10**mu[0], 10**mu[1], mu[2], 10**mu[3], 10**mu[4], mu[5], 
                 mu[6], mu[7], 10**mu[8], mu[9]]
        return start


    def _get_freq_range(self):
        """ Get frequency range for model
        """
        
        dnu, numax, eps = self.start[:3]
        
        nmax = self._get_nmax(dnu, numax, eps)
        enns = self._get_enns(nmax, self.norders)

        lfreq = (min(enns) - 1.25 + eps) * dnu
        ufreq = (max(enns) + 1.25 + eps) * dnu
        return lfreq, ufreq    
    
    
    
    def _start_init(self, verbose=False):
        """ This is in pre-alpha
        
        Bodge a better starting point
        """
        
        like_start = np.ones(len(self.start_samples[:, 0]))
        for idx, samp in enumerate(self.start_samples):
            like_start[idx] = self.likelihood(samp)
        if verbose:
            print(f'Likelihood at the start : {np.max(like_start)}')
            print(f'Start params from init : {self.start_samples[np.argmax(like_start), :]}')



""" Detect l=1 ridges

This is a basic module for detecting the l=1 modes in a spectrum. 

It attempts to find the l=1 ridge by using a clustering function on frequency
bins that are inconsistent with noise at a ~10% probability (H0 test).

It uses the output from the pbjam.peakbag module, where the best-fit models for
the l=2,0 modes is divided out, to form a residual spectrum. This is used in 
the following to detect the l=1 ridge under the assumption that the remaining
significant peaks are most likely caused by l=1 modes (not l=3).

Several levels of binning are applied and the H0 test is performed at each
level. Peaks that persist at many levels of binning will be picked up by 
Hdbscan (unsupervised clustering algorithm), and the average frequencies of the 
clusters are taken to be an estimate of the l=1 frequencies. 

The frequency range between l=0 and l=2, for a given n, is searched for the
cluster with the largest number of samples, and it is assumed that this is
the l=1 ridge. Only 1 l=1 mode is picked up for each radial order, i.e., 
mixed modes are not considered. 

Note
----
It is recommended that pbjam.peakbag be run prior to using this module.

"""

import numpy as np
from scipy.special import gammaincc
from sklearn.preprocessing import MinMaxScaler
from sklearn.utils import shuffle as skshuffle
import hdbscan as Hdbscan
import warnings
from .plotting import plotting
import astropy.units as units
import lightkurve as lk

class ellone(plotting):
    """ Basic l=1 detection 

    Uses either an instance of the pbjam.peakbag module where the l=2,0 modes 
    have been fit, or separate spectrum as inputs to attempt to find the l=1
    modes.
    
    Parameters
    ----------
    pbinst : pbjam.peakbag instance (optional)
        An instance of peakbag where the l=2,0 modes have been fit
    f : ndarray (optional)
        Numpy array of frequency bins of the spectrum (muHz).
    s : ndarray (optional)
        Numpy array of power in each frequency bin (SNR).
    
    Attributes
    ----------
    norders : int
        Number of orders to consider, taken from the peakbag instance
    res : ndarray
        Array of residual power after dividing out the l=2,0 modes
    hdblabels : ndarray
        Array of cluster labelss assigned by HDBscan
    hdbX : ndarray
        Array of input samples to HDBscan
    hdb_clustern : ndarray
        Number of samples retained in each cluster
    
    Note
    ----
    If not pbjam.peakbag instance is supplied, a spectrum must be supplied 
    instead, in which case the l=2,0 modes may be picked up instead of the l=1.
    
    """
    
    def __init__(self, pbinst=None, f=None, s=None):
        
        if pbinst:
            
            if (f is not None) and (s is not None):
                self.f = f#
                self.s = s#
            else:
                self.f = pbinst.f
                self.s = pbinst.s
                
            self.norders = pbinst.norders
            self.pg = pbinst.pg
            self.pbinst = pbinst
            self.res = self.residual()
            
        
        elif (not f is None) and (not s is None):
            self.f = f
            self.s = s
            self.res = s
            self.pg = lk.periodogram.Periodogram(f*units.microhertz,
                                                 units.Quantity(s, None))
        else:
            raise AssertionError('Must provide frequency and spectrum input')
        
        self.hdblabels = None
        self.hdbX = None
        self.hdb_clusterN = None
        
    def residual(self,):
        """ Compute the residual after dividing out l=2,0
        
        Uses the best-fit model from a pbjam.peakbag run and divides out the 
        l=2,0 mode power from the supplied spectrum.
        
        Returns
        -------
        res : ndarray
            Array of residual power after dividing out the l=2,0 modes
            
        """
        
        res = self.s.copy()
        keys = ['l0','l2', 'width0', 'width2', 'height0', 'height2', 'back']
        orders = range(self.norders)
        smry = self.pbinst.summary
        varbs = np.array([[smry.loc[[f'{string}__{i}'], 'mean'].values[0] for i in orders] for string in keys])
        mod = self.pbinst.model(*varbs)
        for i in orders:
            flad = self.pbinst.ladder_f[i,:]
            idx = (flad[0] <= self.f) & (self.f <= flad[-1])
            res[idx] /= mod[i,:]
        return res
     
    def binning(self, nbin):
        """ Simply mean-binning
        
        Bins the spectrum by a factor of nbin
        
        Parameters
        ----------
        nbin : int
            Number of bins to average
        
        Note
        ----
        Assumes the frequency bins are equidistant 
        
        """
        
        nmax = int(len(self.res)/nbin)*nbin
        sbin = self.res[:nmax].reshape((-1, nbin)).mean(axis=1)
        fbin = self.f[:nmax].reshape((-1, nbin)).mean(axis=1)
        return fbin, sbin
    
    def H0test(self, fbin, sbin, nbin, dnu, reject=0.1): 
        """ Perform H0 test on SNR spectrum
        
        Parameters
        ----------
        fbin : ndarray
            Numpy array of frequency bins of the spectrum (muHz), binned by a
            factor of nbin.
        sbin : ndarray
            Numpy array of power in each frequency bin (SNR), binned by a
            factor of nbin. 
        nbin : int
            Binning factor
        dnu : float
            Large separation of the p-modes (muHz)
        reject: float
            Rejection level to use for the H0 test. (~0.1)
            
        """
        
        Nind = int(dnu/(2*nbin*np.median(np.diff(fbin))))
        g = gammaincc(nbin, nbin*sbin)
        k = 1-(1-g)**Nind
        idx = k < reject
        return idx, k
    
    
    def H0_inconsistent(self, dnu, Nmax, rejection_level):
        """ Find bins inconsistent with noise
        
        Perform the H0 test at several degrees of binning in freqeuency. 
        Peaks that are inconsistent with noise are retained. 
        
        Parameters
        ----------
        dnu : float
            Large separation of the p-modes (muHz)
        Nmax : int
            Maximum level of binning to try. Default level might require 
            adjusting for lower resolution spectra.
        rejection_level: float
            Probability level to perform H0 test at. Peaks in the spectrum 
            with a probability of being consistent with noise, that is less
            than this are retained.
        
        Returns
        -------
        nu : ndarray
            Frequency of peaks that satisfy the H0 test.
        N : ndarray
            Bin factors at which the significant peaks were selected.
        pH0s : ndarray
            Probabilities of the peaks that are retained
            
        """
        
        N = np.array([])
        nu = np.array([])
        pH0s = np.array([])
        
        for nbin in range(1, Nmax+1):
            fbin, sbin = self.binning(nbin) 
            idx, ps = self.H0test(fbin, sbin, nbin, dnu, rejection_level) 
            N = np.append(N, np.zeros(len(fbin[idx]))+nbin)
            nu = np.append(nu, fbin[idx])
            pH0s = np.append(pH0s, ps[idx])
            
        return nu, N, pH0s
    
    def clustering_preprocess(self, nu, N, limits = (0, 100000)):
        """ Preprocess the samples before clustering
        
        Preprocesses the list of frequencies at which significant peaks in the 
        power spectrum were found. The binning factors are shuffled to prevent
        clustering along that axis (axis=1). 
        
        The binning factors are scaled to range between 0 and 1. 
        
        Parameters
        ----------
        nu : ndarray
            Frequency of peaks that satisfy the H0 test.
        N : ndarray
            Bin factors at which the significant peaks were selected.
        limits : list
            Lower and upper limits in nu to use for clustering. Samples
            beyond these limits are rejected.
        
        Returns
        -------
        X : ndarray
            Array of samples to be used by HDBscan 
            
        """
        
        nuidx = (limits[0] < nu) & (nu < limits[1])
        Nscaler = MinMaxScaler().fit(N.reshape(-1,1))
        Ns = skshuffle(Nscaler.transform(N[nuidx].reshape(-1,1))).flatten()
        return np.vstack((nu[nuidx], Ns)).T
    
    def span(self, x):
        """ Compute span of array
        
        Parameters
        ----------
        x : ndarray
            List of floats/integers
        
        Returns
        -------
        span : float
            Range spanned by the minimum and maximum values in x
            
        """
        
        return max(x)-min(x)
    
    def clustering(self, nu, N, Nmax, outlier_limit=0.5, cluster_prob=0.9):
        """ Perform HDBscan clustering
        
        Uses HDBscan to perform an unsupervised clustering analysis of the 
        frequencies of the H0 significance test. 
        
        Samples in the clusters are retained based on their probability of 
        being part of an individual cluster, and the odds of them being an 
        outlier of all the samples in the list.
        
        Parameters
        ----------
        nu : ndarray
            Frequency of peaks that satisfy the H0 test.
        N : ndarray
            Bin factors at which the significant peaks were selected.
        Nmax: int
            Maximum binning factor used in the H0 test
        outlier_limiter : float
            Probability that a sample is an outlier of a cluster. Samples with
            a probability above this are rejected.
        cluster_prob : float
            Probability that samples belong to a particular cluster. Samples
            with a probability below this are rejected.
            
        Returns
        -------
        nus : ndarray
            Array of mean frequencies of the clusters identified by HDBscan
        Wratios : ndarray
            Ratio of the span of the cluster and the average mode width of the
            p-modes. (unused)
            
        """
        
        X = self.clustering_preprocess(nu, N)
        hdbscan = Hdbscan.HDBSCAN(min_cluster_size = 2*Nmax).fit(X)
        
        labels = hdbscan.labels_
        ulabels = np.unique(labels)
        
        nus = np.zeros(len(ulabels))
        nstds = np.zeros(len(ulabels))
        
        for i, u in enumerate(ulabels):
            hdbidx = (labels==u) & (hdbscan.outlier_scores_<outlier_limit) & (hdbscan.probabilities_>cluster_prob) 
            if (len(X[hdbidx,0])==0) or (u == -1):
                continue
            nus[i] = np.mean(X[hdbidx,0])
            nstds[i] = np.std(X[hdbidx,0])
            #Wratios[i] = self.span(X[hdbidx,0])/W
    
        self.hdblabels = labels
        self.hdbX = X
        self.hdb_clusterN = np.array([len(labels[labels==i]) for i in ulabels])
        
        return nus[1:], nstds[1:]
    

    
    def get_ell1(self, dnu):
        """ Estimate frequency of l=1 modes (p-modes)
        
        Takes the best-fit l=2,0 mode frequencies from pbjam.peakbag and 
        searches the frequency range in between subsequent radial orders for
        clusters of significant peaks. The cluster with the largest number of 
        peaks is assumed to be the most likely l=1 location. 
        
        Parameters
        ----------
        dnu : float
            Large separation of the p-modes (muHz)
            
        Returns
        -------
        nul1s : ndarray
            Array of estimated frequencies of the l=1 modes. 
            
        Note
        ----
        Raises a warning if the detected frequency is far from the expected 
        frequency based on the Universal Pattern. 
        
        """
        
        pbsmry = self.pbinst.summary
        N = self.norders
        ell02 = ['l0', 'l2']
        nul0s, nul2s = np.array([[pbsmry.loc[f'{x}__{i}', 'mean'] for i in range(N)] for x in ell02])

        d01 = (1./2 -0.0056 -0.002*np.log10(dnu))*dnu
        
        nul1s = np.zeros(len(nul0s))
        nul1s_std = np.zeros(len(nul0s))
        
        for i in range(len(nul0s)):
            nuidx = (nul0s[i] < self.cluster_means) & (self.cluster_means < nul2s[i]+dnu)
            
            if len(self.hdb_clusterN[1:][nuidx]) == 0:
                continue
            
            maxidx = np.argmax(self.hdb_clusterN[1:][nuidx])
            nul1s[i] = self.cluster_means[nuidx][maxidx]
            nul1s_std[i] = self.cluster_stds[nuidx][maxidx]
            
            if (nul0s[i] - nul1s[i])/d01 > 0.2:
                warnings.warn('Cluster nu_l1 exceeds UP estimate by more than 20%')
                
        return nul1s, nul1s_std

    def __call__(self, dnu, Nmax = 30, rejection_level = 0.1):
        """ Perform all the steps to estimate l=1 frequencies
        
        Check which peaks in the provided spectrum are inconsistent with the
        noise level, at various degrees of binning.
        
        Peaks that are persistent over many binning levels are clustered and 
        the mean frequencies are assumed to be likely locations of the l=1.
        modes.
        
        Parameters
        ----------
        dnu : float
            Large separation of the p-modes (muHz)
        Nmax : int (optional)
            Maximum level of binning to try. Default level might require 
            adjusting for lower resolution spectra.
        rejection_level: float (optional)
            Probability level to perform H0 test at. Peaks in the spectrum 
            with a probability of being consistent with noise, that is less
            than this are retained.
        
        Returns
        -------
        nul1s : ndarray
            Array of estimated frequencies of the l=1 modes. 
        
        """
        
        nus, counts, pH0s = self.H0_inconsistent(dnu, Nmax, rejection_level)
        self.cluster_means, self.cluster_stds = self.clustering(nus, counts, Nmax)
        self.nu_l1, self.nu_l1_std = self.get_ell1(dnu)
        return self.nu_l1
    
         
# d02 = 10**st.asy_fit.summary.loc['d02', '50th']


from . import PACKAGEDIR
import os
import numpy as np

def get_priorpath():
    """ Get default prior path name
    
    Returns
    -------
    priorpath : str
        Default path to the prior in the pacakage directory structure
        
    """
    
    return os.path.join(*[PACKAGEDIR, 'data', 'prior_data.csv'])


def get_percentiles(X, sigma = 2, **kwargs):
    """ Get percentiles of an distribution
    
    Compute the percentiles corresponding to sigma=1,2,3.. including the 
    median (50th), of an array.
    
    Parameters
    ----------
    X : numpy.array()
        Array to find percentiles of
    sigma : int
        Sigma values to compute the percentiles of 68% 95% etc.
    kwargs : dict
        Arguments to be passed to numpy.percentile
    
    returns
    -------
    percentiles : numpy.array()
        Numpy array of percentile values of X.
    
    """
    
    percs = np.array([0.682689492137,
                      0.954499736104,
                      0.997300203937,
                      0.999936657516,
                      0.999999426697,
                      0.999999998027])*100/2    
    percs = np.append(0, percs)    
    percs = np.append(-percs[::-1][:-1],percs)
    percs += 50
    
    return np.percentile(X, percs[6-sigma : 6+sigma+1], **kwargs)


def to_log10(x, xerr):
    """ Transforms observables into log10 space.

    """

    if xerr > 0:
        return [np.log10(x), xerr/x/np.log(10.0)]
    return [x, xerr]

def normal(x, mu, sigma):
    """ logarithm of normal distribution (not normalized!!)

    Evaluates the logarithm of a normal distribution at x. 

    Inputs
    ------
    x : real
        observed value
    mu : real
        distribution mean
    sigma : real
        distribution standard deviation

    Returns
    -------
    y : real
        Logarithm of the normal distribution at x
    """

    if (sigma < 0):
        return 0.0
    return -0.5 * (x - mu)**2 / sigma**2


import emcee
import numpy as np
import scipy.stats as st

class mcmc():
    """ Class for MCMC sampling

    Use EMCEE to fit a provided model to a spectrum.

    Parameters
    ----------
    start : ndarray
        An array of size ndim for the starting position of the parameters.
    likelihood : function
        Function to call that returns the log likelihood when passed the
        parameters.
    lp : function
        Function to call that returns the log prior probability when passed
        the parameters.
    nwalkers : int
        The number of walkers that emcee will use.

    """

    def __init__(self, start, likelihood, prior, nwalkers=50):

        self.start = start
        self.ndim = len(start)
        self.likelihood = likelihood
        self.prior = prior

        self.nwalkers = nwalkers

        self.sampler = emcee.EnsembleSampler(self.nwalkers,
                                             self.ndim,
                                             self.logpost)

        self.chain = None
        self.flatchain = None
        self.lnlike = None
        self.flatlnlike = None
        self.acceptance = None

    def logpost(self, p):
        """ Evaluate the likelihood and prior
        
        Returns the log posterior probability given parameters p. Evaluates
        first the prior function and then the likelihood function. In the 
        event that the prior returns -inf, the function exits.
        
        Parameters
        ----------
        p : list
            Fit parameters
        
        Returns
        -------
        log_posterior: float
            log posterior of the model given parameters p and the observed
            quantities.
        
        """
        
        logp = self.prior(p)
        if logp == -np.inf:
            return -np.inf

        loglike = self.likelihood(p)
        return logp + loglike

    def stationarity(self, nfactor=20):
        """
        Tests to see if stationarity metrics are satified.

        nfactor should be nearer 100 follwing the emcee docs
        but in PBjam useage I get better than reasonable results
        when using 20.
        
        """
        
        tau = self.sampler.get_autocorr_time(tol=0)
        converged = np.all(tau * nfactor < self.sampler.iteration)
        return converged


    def __call__(self, max_iter=20000, spread=1e-4, start_samples=[]):
        """ Initialize and run the EMCEE afine invariant sampler

        Parameters
        ----------
        max_iter: int
            Don't run the sampler for longer than this.  The sampler will
            stop if convergence is deemed to have been achieved.
        spread : float
            Percent spread around the intial position of the walkers. Small
            value starts the walkers in a tight ball, large value fills out
            the range set by parameter bounds.
        start_samples: ndarray
            An array that has samples from the distribution that you want to
            start the sampler at.

        Returns
        -------
        sampler.flatchain : array
            The chain of (nwalkers, niter, ndim) flattened to
            (nwalkers*niter, ndim).
            
        """
        
        nsteps = 1000

        # Start walkers in a tight random ball
        if len(start_samples) == 0:
            # Do this in the case of KDE
            pos = np.array([self.start + (np.random.randn(self.ndim) * spread) for i in range(self.nwalkers)])
        else:
            # Do this in the case of Asy_peakbag, should be replaced with the actual sample
            pos = np.random.randn(self.nwalkers, self.ndim)
            pos *= start_samples.std(axis=0)
            pos += start_samples.mean(axis=0)

        # Burn in
        pos, prob, state = self.sampler.run_mcmc(initial_state=pos, nsteps=nsteps)
        # Fold in low AR chains
        pos = self.fold(pos, spread=spread)
        # Reset sampler
        self.sampler.reset()

        # Run with burnt-in positions
        pos, prob, state = self.sampler.run_mcmc(initial_state=pos, nsteps=nsteps)
        while not self.stationarity():
            pos, prob, state = self.sampler.run_mcmc(initial_state=pos, nsteps=nsteps)
            print(f'Steps taken: {self.sampler.iteration}')
            if self.sampler.iteration == max_iter:
                break
        if self.sampler.iteration < max_iter:
            print(f'Chains reached stationary state after {self.sampler.iteration} iterations.')
        elif self.sampler.iteration == max_iter:
            print(f'Sampler stopped at {max_iter} (maximum). Chains did not necessarily reach a stationary state.')
        else:
            print('Unhandled exception')

        # Fold in low AR chains and run a little bit to update emcee
        self.fold(pos, spread=spread)
        pos, prob, state = self.sampler.run_mcmc(initial_state=pos, nsteps=100, store=True)

        # Final acceptance
        self.acceptance = self.sampler.acceptance_fraction

        # Estimate autocorrelation time
        tau = self.sampler.get_autocorr_time(tol=0, discard = nsteps).mean()

        # 3D chains
        discard = int(tau*5)
        thin = int(tau/4)
        self.chain = self.sampler.get_chain(discard=discard, thin=thin, flat=False)
        self.lnlike = self.sampler.get_log_prob(discard=discard, thin=thin, flat=False)

        # 2D chains
        self.flatchain = self.sampler.get_chain(discard=discard, thin=thin, flat=True)
        self.flatlnlike = self.sampler.get_log_prob(discard=discard, thin=thin, flat=True)

        self.sampler.reset()  # This hopefully minimizes emcee memory leak
        return self.flatchain


    def fold(self, pos, accept_lim = 0.2, spread=0.1):
        """ Fold low acceptance walkers into main distribution

        At the end of the burn-in, some walkers appear stuck with low
        acceptance fraction. These can be selected using a threshold, and
        folded back into the main distribution, estimated based on the median
        of the walkers with an acceptance fraction above the threshold.

        The stuck walkers are relocated with multivariate Gaussian, with mean
        equal to the median of the high acceptancew walkers, and a standard
        deviation equal to the median absolute deviation of these.

        Parameters
        ----------
        pos : array
            The final position of the walkers after the burn-in phase.
        accept_lim: float
            The value below which walkers will be labelled as bad and/or hence
            stuck.
            
        """
        
        idx = self.sampler.acceptance_fraction < accept_lim
        nbad = np.shape(pos[idx, :])[0]
        if nbad > 0:
            flatchains = self.sampler.chain[~idx, :, :].reshape((-1, self.ndim))
            good_med = np.median(flatchains, axis = 0)
            good_mad = st.median_absolute_deviation(flatchains, axis = 0) * spread
            pos[idx, :] = np.array([np.random.randn(self.ndim) * good_mad + good_med for n in range(nbad)])
        return pos



import numpy as np
import pandas as pd
from .mcmc import mcmc
import warnings
from .plotting import plotting
import statsmodels.api as sm
from .jar import get_priorpath, to_log10, normal

class kde(plotting):
    """ A class to predict epsilon.

    TODO: See the docs for full information. (especially example_advanced)

    """

    def __init__(self, starinst=None, prior_file=None):

        if starinst:
            self.f = starinst.f
            self.s = starinst.s
            self.pg = starinst.pg
            starinst.kde = self

            if prior_file is None:
                prior_file = starinst.prior_file
        elif prior_file is None:
            prior_file = get_priorpath()

        self.verbose = False
        self.prior_data = pd.read_csv(prior_file)

    def select_prior_data(self, numax=None, KDEsize = 100):
        """ Selects useful prior data based on proximity to estimated numax.

        Selects a subset of targets around input numax to use for computing the
        KDE. If the number of targets in the range considered for the prior is
        less than 100, the range will be expanded until it ~100. This is to
        ensure that the KDE can be constructed. If the initial range includes
        more than 100 targets (e.g., in dense parts of the RGB) the selection
        of targets will be downsampled randomly.


        Note: does not ensure that the KDE is finite at the location of your
        target

        Parameters
        ----------

        numax : list
            The estimate of numax together with uncertainty in log space.
            If numax==None then no selection will be made - all data will
            be used (you probably don't want to do this).

        KDEsize : int
            Number of targets to include in the KDE estimation.

        """

        if not numax:
            return self.prior_data

        # Select numax range and expand if needed
        idx = self._prior_expand_check(numax, KDEsize)

        # Downsample to KDEsize
        self.prior_data = self._downsample_prior_check(idx, KDEsize)


    def _downsample_prior_check(self, idx, KDEsize):
        """ Reduce prior sample size

        Randomly picks KDEsize targets from the prior sample

        Parameters
        ----------
        idx : boolean array
            Array mask to select targets to be included in the KDE estimation.

        KDEsize : int
            Number of targets to include in the KDE estimation.

        """
        if len(idx) < KDEsize:
            warnings.warn(f'Sample for estimating KDE is less than the request {KDEsize}.')
            KDEsize = len(idx)

        return self.prior_data.sample(KDEsize, weights = idx, replace = False)

    def _prior_expand_check(self, numax, KDEsize):
        """ Expand numax interval to reach sufficient KDE sample size

        If necessary, increases the range around starting numax to use as the
        initial prior, until the sample contains at least N = KDEsize targets.

        Parameters
        ----------
        numax : length 2 list [numax, numax_err]
            The estimate of numax together with uncertainty in log space.

        KDEsize : int
            Number of targets to include in the KDE estimation.

        Returns
        -------
        idx : boolean array
            Array mask to select targets to be included in the KDE estimation.

        """

        nsigma = 1

        idx = np.abs(self.prior_data.numax.values - numax[0]) < nsigma * numax[1]

        flag_warn = False
        while len(self.prior_data[idx]) < KDEsize:
            idx = np.abs(self.prior_data.numax.values - numax[0]) < nsigma * numax[1]

            if not flag_warn:
                warnings.warn(f'Only {len(self.prior_data[idx])} star(s) near provided numax. ' +
                'Expanding the range to include ~100 stars.')
                flag_warn = True

            if nsigma >= KDEsize:
                break

            nsigma += 0.1

        return idx



    def make_kde(self, bw_fac=1.0):
        """ Takes the prior data and constructs a KDE function

        TODO: add details on the band width determination - see example
        advanced for a base explaination.

        I have settled on using the values from a cross validated maximum
        likelihood estimate.

        """

        self.par_names = ['dnu', 'numax', 'eps', 'd02', 'alpha', 'env_height',
                          'env_width', 'mode_width', 'teff', 'bp_rp']

        self.select_prior_data(self.log_obs['numax'])

        if self.verbose:
                print(f'Selected data set length {len(self.prior_data)}')

        if bw_fac != 1:
            from statsmodels.nonparametric.bandwidths import select_bandwidth
            bw = select_bandwidth(self.prior_data[self.par_names].values,
                                  bw = 'scott',
                                  kernel=None) * bw_fac

        else:
            if self.verbose:
                print('Selecting sensible stars for kde')
                print(f'Full data set length {len(self.prior_data)}')
            bw = 'cv_ml'

        self.kde = sm.nonparametric.KDEMultivariate(
                            data=self.prior_data[self.par_names].values,
                            var_type='cccccccccc', bw=bw)



    def prior(self, p):
        """ Calculates the log prior

        Evaluates the KDE for the parameters p. Additional hard/soft priors
        can be added here as needed to, e.g., apply boundaries to the fit.

        Hard constraints should be applied at the top so function exits early,
        if necessary.

        Parameters
        ----------
        p : array
            Array of model parameters

        Returns
        -------
        lp : real
            The log likelihood evaluated at p.

        """

        # d02/dnu < 0.2  (np.log10(0.2) ~ -0.7)
        if p[3] - p[0] > -0.7:
            return -np.inf

        # Constraint from prior
        lp = np.log(self.kde.pdf(p))

        return lp

    def likelihood(self, p):
        """ Calculate likelihood

        Calculates the likelihood of the observed properties given the proposed
        parameters p.

        Parameters
        ----------
        p : array
            Array of model parameters

        Returns
        -------
        lnlike : float
            The log likelihood evaluated at p.

        """

        lnlike = 0.0

        # Constraints from observational data
        lnlike += normal(p[0], *self.log_obs['dnu'])
        lnlike += normal(p[1], *self.log_obs['numax'])
        lnlike += normal(p[8], *self.log_obs['teff'])
        lnlike += normal(p[9], *self.obs['bp_rp'])

        return lnlike

    def kde_sampler(self, nwalkers=50):
        """ Samples from the posterior probability distribution

        p(theta | D) propto p(theta) p(D | theta)

        p(theta) is given by the KDE function of the prior data

        p(D | theta) is given by the observable constraints

        Samples are drawn using `emcee`.

        Returns
        -------

        chains: flatchain
            The emcee chain of samples from the posterior

        Note:['dnu', 'numax', 'eps',
                     'd02', 'alpha', 'env_height',
                     'env_width', 'mode_width', 'teff',
                     'bp_rp']
        key: [log, log, lin, log, log, log, log, log, log, lin]

        """

        if self.verbose:
            print('Running KDE sampler')

        x0 = [self.log_obs['dnu'][0],  # log10 dnu
              self.log_obs['numax'][0],  # log10 numax
              1.0,  # eps
              np.log10(0.1 * self.obs['dnu'][0]),  # log10 d02
              -2.0,  # log10 alpha
              1.0,  # log10 env height
              1.0,  # log10 env width,
              -1.0,  # log10 mode width
              self.log_obs['teff'][0],
              self.obs['bp_rp'][0]]

        self.fit = mcmc(x0, self.likelihood, self.prior, nwalkers=nwalkers)

        return self.fit()


    def kde_predict(self, n):
        """
        Predict the frequencies from the kde method samples.

        The sampler must be run before calling the predict method.

        Parameters
        ----------

        n: numpy_array
            The radial order

        Returns
        -------
        frequencies: numpy-array
            A numpy array of length len(n) containting the frequency estimates.

        frequencies_unc: numpy-array
            A numpy array of length len(n) containting the frequency estimates
            uncertainties.

        """

        if not hasattr(self, 'samples'):
            print('Need to run the sampler first')
            return -1, -1
        dnu = 10**self.samples[:, 0]
        eps = self.samples[:, 2]
        nmax = 10**self.samples[:, 1] / dnu - eps
        alpha = 10**self.samples[:, 4]
        freq = np.array([(nn + eps + alpha/2.0 * (nn - nmax)**2) * dnu for nn in n])
        return freq.mean(axis=1), freq.std(axis=1)

    def __call__(self, dnu=[1, -1], numax=[1, -1], teff=[1, -1],
                 bp_rp=[1, -1], verbose=False, bw_fac=1):
        """  Compute and sample the KDE.

        Inputs
        ------
        dnu : [real, real]
            Large frequency spacing and uncertainty
        numax : [real, real]
            Frequency of maximum power and uncertainty
        teff : [real, real]
            Stellar effective temperature and uncertainty
        bp_rp : [real, real]
            The Gaia Gbp - Grp color value and uncertainty
            (probably ~< 0.01 dex)
        Returns
        -------
        result : array-like
            [estimate of epsilon, unceritainty on estimate]

        """
        self.verbose = verbose

        self.obs = {'dnu': dnu, 'numax': numax, 'teff': teff, 'bp_rp': bp_rp}

        self.log_obs = {x: to_log10(*self.obs[x]) for x in self.obs.keys() if x != 'bp_rp'}

        self.make_kde(bw_fac)

        self.samples = self.kde_sampler()

        self.result = [self.samples[:, 2].mean(), self.samples[:, 2].std()]



import os
from .asy_peakbag import asymptotic_fit
from .priors import kde
from .peakbag import peakbag
from .jar import get_priorpath
from .plotting import plotting
import pandas as pd

class star(plotting):
    """ Class for each star to be peakbagged

    Additional attributes are added for each step of the peakbagging process

    Note spectrum is flattened (background divided out.)

    Parameters
    ----------
    ID : string, int
        Target identifier. If custom timeseries/periodogram is provided, it
        must be resolvable by LightKurve (KIC, TIC, EPIC, HD, etc.).

    pg : lightkurve.periodogram.Periodogram object
        A lightkurve periodogram object containing frequencies in units of
        microhertz and power (in arbitrary units).

    numax : list
        List of the form [numax, numax_error]. For multiple targets, use a list
        of lists.

    dnu : list
        List of the form [dnu, dnu_error]. For multiple targets, use a list
        of lists.

    teff : list
        List of the form [teff, teff_error]. For multiple targets, use a list
        of lists.

    bp_rp : list
        List of the form [bp_rp, bp_rp_error]. For multiple targets, use a list
        of lists.

    store_chains : bool, optional
        Flag for storing all the full set of samples from the MCMC run.
        Warning, if running multiple targets, make sure you have enough memory.

    nthreads : int, optional
        Number of multiprocessing threads to use to perform the fit. For long
        cadence data 1 is best, more will just add parallelization overhead.
        Untested on short cadence.

    make_plots : bool, optional
        If True, will save figures when calling methods in `star`.

    path : str, optional
        The path at which to store output. If no path is set but make_plots is
        True, output will be saved in the current working directory.

    verbose : bool, optional
        If True, will show error messages on the users terminal if they occur.

    Attributes
    ----------
    f : array
        Array of power spectrum frequencies
    s : array
        power spectrum
    data_file : str
        Path to the csv file containing the prior data

    """

    def __init__(self, ID, pg, numax, dnu, teff, bp_rp, path=None,
                 prior_file = None):

        self.ID = ID
        self.pg = pg.flatten() # in case user supplies unormalized spectrum
        self.f = self.pg.frequency.value
        self.s = self.pg.power.value

        self.numax = numax
        self.dnu = dnu
        self.teff = teff
        self.bp_rp = bp_rp

        self._set_path(path)
        self._make_output_dir()

        if prior_file is None:
            self.prior_file = get_priorpath()
        else:
            self.prior_file = prior_file

    def _set_path(self, path):
        """ Sets the path attribute for star
        
        If path is a string it is assumed to be a path name, if not the 
        current working directory will be used. 
        
        Parameters
        ----------
        path : str
            Directory to store peakbagging output.
        
        """

        if isinstance(path, str):
            # If path is str, presume user wants to override self.path
            self.path = os.path.join(*[path, f'{self.ID}'])
        else:
            self.path = os.path.join(*[os.getcwd(), f'{self.ID}'])
            
    def _make_output_dir(self):
        """ Make output directory for star

        Attempts to create an output directory for all the results that PBjam
        produces. A directory is created when a star class instance is
        initialized, so a session might create multiple directories.

        """

        # Check if self.path exists, if not try to create it
        if not os.path.isdir(self.path):
            try:
                os.makedirs(self.path)
            except Exception as ex:
                message = "Star {0} produced an exception of type {1} occurred. Arguments:\n{2!r}".format(self.ID, type(ex).__name__, ex.args)
                print(message)

    def run_kde(self, bw_fac=1.0, make_plots=False):
        """ Run all steps involving KDE.

        Starts by creating a KDE based on the prior data sample. Then samples
        this KDE for initial starting positions for asy_peakbag.

        Also generates plots of the results and stores them in the star
        directory.

        Parameters
        ----------
        bw_fac : float
            Scaling factor for the KDE bandwidth. The bandwidth is
            automatically, but may be scaled to adjust for, .e.g, sparsity of
            the prior sample.
        make_plots : bool
            Whether or not to produce plots of the results.

        """

        print('Starting KDE estimation')
        # Init
        kde(self)

        # Call
        self.kde(dnu=self.dnu, numax=self.numax, teff=self.teff,
                 bp_rp=self.bp_rp, bw_fac=bw_fac)

        # Store
        if make_plots:
            self.kde.plot_corner(path=self.path, ID=self.ID,
                                 savefig=make_plots)
            self.kde.plot_spectrum(pg=self.pg, path=self.path, ID=self.ID,
                                   savefig=make_plots)


    def run_asy_peakbag(self, norders=None, make_plots=False,
                        store_chains=False):
        """ Run all stesps involving asy_peakbag.

        Performs a fit of the asymptotic relation to the spectrum (l=2,0 only),
        and outputs result plots and a summary of the fit results.

        Parameters
        ----------
        norders : int
            Number of orders to include in the fits
        make_plots : bool
            Whether or not to produce plots of the results.
        store_chains : bool
            Whether or not to store MCMC chains on disk.

        """

        print('Starting asymptotic peakbagging')
        # Init
        asymptotic_fit(self, norders=norders)

        # Call
        self.asy_fit()

        # Store
        outpath = lambda x: os.path.join(*[self.path, x])
        self.asy_fit.summary.to_csv(outpath(f'asy_fit_summary_{self.ID}.csv'),
                                    index=True, index_label='name')
        self.asy_fit.modeID.to_csv(outpath(f'asy_fit_modeID_{self.ID}.csv'),
                                   index=False)

        if make_plots:
            self.asy_fit.plot_spectrum(path=self.path, ID=self.ID,
                                       savefig=make_plots)
            self.asy_fit.plot_corner(path=self.path, ID=self.ID,
                                       savefig=make_plots)

        if store_chains:
            pd.DataFrame(self.asy_fit.samples, columns=self.asy_fit.par_names).to_csv(outpath(f'asy_peakbag_chains_{self.ID}.csv'), index=False)



    def run_peakbag(self, model_type='simple', tune=1500, nthreads=1,
                    make_plots=False, store_chains=False):
        """  Run all stesps involving peakbag.

        Performs fit using simple lorentzian pairs two subsections of the
        power spectrum based on results from asy_peakbag.

        Parameters
        ----------
        model_type : str
            Defaults to 'simple'.
            Can be either 'simple' or 'model_gp' which sets the type of model
            to be fitted to the data.
        tune : int
            Numer of tuning steps passed to pm.sample
        make_plots : bool
            Whether or not to produce plots of the results.
        store_chains : bool
            Whether or not to store MCMC chains on disk.
        nthreads : int
            Number of processes to spin up in pymc3

        """

        print('Starting peakbagging')
        # Init
        peakbag(self, self.asy_fit)

        # Call
        self.peakbag(model_type=model_type, tune=tune, nthreads=nthreads)

        # Store
        outpath = lambda x: os.path.join(*[self.path, x])
        self.peakbag.summary.to_csv(outpath(f'peakbag_summary_{self.ID}.csv'),
                                    index_label='name')

        if store_chains:
            pass # TODO need to pickle the samples if requested.
        if make_plots:
            self.peakbag.plot_spectrum(path=self.path, ID=self.ID,
                                       savefig=make_plots)


    def __call__(self, bw_fac=1.0, norders=8, model_type='simple', tune=1500,
                 verbose=False, make_plots=True, store_chains=True, nthreads=1):
        """ Perform all the PBjam steps


        Parameters
        ----------
        bw_fac : float
            Scaling factor for the KDE bandwidth. The bandwidth is
            automatically, but may be scaled to adjust for, .e.g, sparsity of
            the prior sample.
        norders : int
            Number of orders to include in the fits
        model_type : str
            Defaults to 'simple'. Can be either 'simple' or 'model_gp' which
            sets the type of model to be fit for the mode linewidths.
        tune : int
            Numer of tuning steps passed to pm.sample
        verbose : bool
            Should I say anything?
        make_plots : bool
            Whether or not to produce plots of the results.
        store_chains : bool
            Whether or not to store MCMC chains on disk.

        """

        self.run_kde(bw_fac=bw_fac, make_plots=make_plots)

        self.run_asy_peakbag(norders=norders, make_plots=make_plots,
                             store_chains=store_chains)

        self.run_peakbag(model_type=model_type, tune=tune, nthreads=nthreads,
                         make_plots=make_plots, store_chains=store_chains)



""" Module for a general set of plotting methods

This module contains a set of plotting methods which are inherited by the
different classes of PBjam.

Contains:
- Spectrum plot, including a best-fit model if available
- Corner plot, showing the marginalized posteriors of each fit parameter (can become quite big for peakbag stage)
- Echelle diagram

As each step is completed, the resulting class instance will inherit the
available plotting methods.

"""

import matplotlib.pyplot as plt
import astropy.convolution as conv
import pbjam, os, corner, warnings
import numpy as np
from pymc3.gp.util import plot_gp_dist
import astropy.units as u
import pymc3 as pm

class plotting():

    def __init__(self):
        pass

    def plot_echelle(self, pg=None):
        """

        Plots an echelle diagram with mode frequencies if available.

        Parameters
        ----------
        pg : periodogram
            A lightkurve periodogram

        Returns
        -------
        fig : figure
            Matplotlib figure object

        """

        freqs = {'l'+str(i): {'nu': [], 'err': []} for i in range(4)}

        if type(self) == pbjam.star:
            dnu = self.dnu[0]
            numax = self.numax[0]

        elif type(self) == pbjam.priors.kde:
            dnu = 10**np.median(self.samples[:,0])
            numax = 10**np.median(self.samples[:,1])

        elif type(self) == pbjam.asy_peakbag.asymptotic_fit:
            dnu = 10**self.summary.loc['dnu', '50th']
            numax = 10**self.summary.loc['numax', '50th']
            for l in np.arange(4):
                idx = self.modeID.ell == l
                freqs['l'+str(l)]['nu'] = self.modeID.loc[idx, 'nu_med']
                freqs['l'+str(l)]['err'] = self.modeID.loc[idx, 'nu_mad']

        elif type(self) == pbjam.peakbag:
            numax = 10**self.asy_result.summary.loc['numax', '50th']
            for l in np.arange(4):
                ell = 'l'+str(l)
                freqs[ell]['nu'] = self.summary.filter(like=ell, axis=0).loc[:, 'mean']
                freqs[ell]['err'] = self.summary.filter(like=ell, axis=0).loc[:, 'sd']
            dnu = np.median(np.diff(freqs['l0']['nu']))

        elif type(self) == pbjam.ellone:
            numax = 10**self.pbinst.asy_result.summary.loc['numax', '50th']
            for l in [0, 2]:
                ell = 'l'+str(l)
                freqs[ell]['nu'] = self.pbinst.summary.filter(like=ell, axis=0).loc[:, 'mean']
                freqs[ell]['err'] = self.pbinst.summary.filter(like=ell, axis=0).loc[:, 'sd']
            freqs['l1']['nu'] = self.nu_l1
            freqs['l1']['err'] = self.nu_l1_std

            dnu = np.median(np.diff(freqs['l0']['nu']))


        else:
            raise ValueError('Unrecognized class type')

        # make dnu an intger multiple of bw
        dnu -= dnu % (self.f[1] - self.f[0])
        nmin = np.floor(self.f.min() / dnu) + 1

        if pg:
            peri = pg
        elif hasattr(self, 'pg'):
            peri = self.pg
        else:
            raise ValueError('Need spectrum to plot echelle diagram')

        seismology = peri.flatten().to_seismology()

        fig, ax = plt.subplots(figsize=[16,9])
        ax = seismology.plot_echelle(deltanu=dnu * u.uHz,
                                     numax=numax * u.uHz,
                                     ax=ax)

        # Overplot modes
        cols = ['C1', 'C2', 'C3', 'C4']

        for l in np.arange(4):
            ell = 'l'+str(l)
            if len(freqs[ell]['nu']) > 0:
                nu = freqs[ell]['nu']
                err = freqs[ell]['err']
                ax.errorbar(nu%dnu, (nu//dnu) * dnu, xerr=err, fmt='o', color = cols[l], label = r'$\ell=$%i' % (l))
        ax.legend(fontsize = 'x-small')

        return fig

    def plot_corner(self, path=None, ID=None, savefig=False):

        """
        Makes a nice corner plot of the fit parameters

        Parameters
        ----------
        path : str (optional)
            Used along with savefig, sets the output directory to store the
            figure.
        ID : str (optional)
            ID of the target to be included in the filename of the figure.
        savefig : bool
            Whether or not to save the figure to disk

        Returns
        -------
        fig : object
            Matplotlib figure object

        """

        if not hasattr(self, 'samples'):
            warnings.warn(f"'{self.__class__.__name__}' has no attribute 'samples'. Can't plot a corner plot.")
            return None

        fig = corner.corner(self.samples, labels=self.par_names,
                            show_titles=True, quantiles=[0.16, 0.5, 0.84],
                            title_kwargs={"fontsize": 12})

        # TODO there should be a check if path is full filepath or just dir
        if path and ID:
            outpath = os.path.join(*[path,  type(self).__name__+ '_corner_' + str(ID) + '.png'])
            if savefig:
                fig.savefig(outpath)

        return fig


    def plot_spectrum(self, pg=None, path=None, ID=None, savefig=False):
        """ Plot the power spectrum

        Plot the power spectrum around the p-mode envelope. Calling this
        method from the different classes such as KDE or peakbag, will plot
        the relevant result from those classes if available.

        Parameters
        ----------
        pg : periodogram
            A lightkurve periodogram
        path : str (optional)
            Used along with savefig, sets the output directory to store the
            figure.
        ID : str (optional)
            ID of the target to be included in the filename of the figure.
        savefig : bool
            Whether or not to save the figure to disk

        Returns
        -------
        fig : object
            Matplotlib figure object

        """

        if not pg and hasattr(self, 'pg'):
            pg = self.pg

        if pg:
            f = pg.frequency.value
            s = pg.power.value

        elif hasattr(self, 'f') & hasattr(self, 's'):
            f = self.f
            s = self.s
        else:
            raise ValueError('Unable to plot spectrum.')

        # The raw and smoothed spectrum will always be plotted
        fig, ax = plt.subplots(figsize=[16,9])
        ax.plot(f, s, 'k-', label='Data', alpha=0.2)
        fac = 0.1 / (f[1] - f[0]) #0.005 * self.dnu[0]  / (f[1] - f[0])
        kernel = conv.Gaussian1DKernel(stddev=fac)
        smoo = conv.convolve(s, kernel)
        ax.plot(f, smoo, 'k-', label='Smoothed', lw=3, alpha=0.6)

        # Overplot kde diagnostic
        if type(self) == pbjam.star:
             xlim = [self.numax[0]-5*self.dnu[0], self.numax[0]+5*self.dnu[0]]

        elif type(self) == pbjam.priors.kde:
            h = max(smoo)
            dnu = 10**(np.median(self.samples[:, 0]))
            nmin = np.floor(min(f) / dnu)
            nmax = np.floor(max(f) / dnu)
            enns = np.arange(nmin-1, nmax+1, 1)
            freq, freq_sigma = self.kde_predict(enns)
            y = np.zeros(len(f))
            for i in range(len(enns)):
                y += 0.8 * h * np.exp(-0.5 * (freq[i] - f)**2 / freq_sigma[i]**2)
            ax.fill_between(f, y, alpha=0.3, facecolor='navy', edgecolor='none',
                            label=r'$\propto P(\nu_{\ell=0})$')

            xlim = [min(freq)-dnu, max(freq)+dnu]

        # Overplot asy_peakbag diagnostic
        elif type(self) == pbjam.asy_peakbag.asymptotic_fit:
            for j in np.arange(-50,0):
                if j==-1:
                    label='Model'
                else:
                    label=None
                ax.plot(f[self.sel], self.model(self.samples[j, :]), 'r-',
                        alpha=0.1)
            for nu in self.modeID['nu_med']:
                ax.axvline(nu, c='k', linestyle='--')
            dnu = 10**self.summary.loc['dnu', '50th']
            xlim = [min(f[self.sel])-dnu,
                    max(f[self.sel])+dnu]

        # Overplot peakbag diagnostic
        elif type(self) == pbjam.peakbag:
            n = self.ladder_s.shape[0]
            par_names = ['l0', 'l2', 'width0', 'width2', 'height0', 'height2',
                         'back']
            for i in range(n):
                for j in range(-50, 0):
                    if (i == 0) and (j==-1):
                        label='Model'
                    else:
                        label=None
                    mod = self.model(*[self.samples[x][j] for x in par_names])
                    ax.plot(self.ladder_f[i, :], mod[i, :], c='r', alpha=0.1,
                            label=label)

            dnu = 10**self.asy_result.summary.loc['dnu', '50th']
            xlim = [min(f[self.asy_result.sel])-dnu,
                    max(f[self.asy_result.sel])+dnu]

        elif type(self) == pbjam.ellone:
            n = self.pbinst.ladder_s.shape[0]
            par_names = ['l0', 'l2', 'width0', 'width2', 'height0', 'height2',
                         'back']
            for i in range(n):
                for j in range(-50, 0):
                    if (i == 0) and (j==-1):
                        mlabel='Model'
                        l1label = r'$\nu_{l=1}$'
                    else:
                        mlabel=None
                    mod = self.pbinst.model(*[self.pbinst.samples[x][j] for x in par_names])
                    ax.plot(self.pbinst.ladder_f[i, :], mod[i, :], c='r', alpha=0.1,
                            label=mlabel)

                ax.axvline(self.nu_l1, color = 'k', alpha = 0.5, label = l1label)

            dnu = 10**self.pbinst.asy_result.summary.loc['dnu', '50th']

            xlim = [min(f[self.pbinst.asy_result.sel])-dnu,
                    max(f[self.pbinst.asy_result.sel])+dnu]

        else:
            raise ValueError('Unrecognized class type')

        ax.set_ylim([0, smoo.max()*1.5])
        ax.set_xlim([max([min(f), xlim[0]]), min([max(f), xlim[1]])])
        ax.set_xlabel(r'Frequency ($\mu \rm Hz$)')
        ax.set_ylabel(r'SNR')
        ax.legend(loc=1)

        if savefig:
            outpath = os.path.join(*[path, f'{type(self).__name__}_{str(ID)}.png'])
            fig.savefig(outpath)

        return fig

    # Asy_peakbag
    def plot_start(self):
        '''
        Plots the starting model as a diagnotstic.
        '''

        dnu = 10**np.median(self.start_samples, axis=0)[0]
        xlim = [min(self.f[self.sel])-dnu, max(self.f[self.sel])+dnu]
        fig, ax = plt.subplots(figsize=[16,9])
        ax.plot(self.f, self.s, 'k-', label='Data', alpha=0.2)
        smoo = dnu * 0.005 / (self.f[1] - self.f[0])
        kernel = conv.Gaussian1DKernel(stddev=smoo)
        smoothed = conv.convolve(self.s, kernel)
        ax.plot(self.f, smoothed, 'k-', label='Smoothed', lw=3, alpha=0.6)
        ax.plot(self.f[self.sel], self.model(self.start_samples.mean(axis=0)),
                'r-', label='Start model', alpha=0.7)
        ax.set_ylim([0, smoothed.max()*1.5])
        ax.set_xlim(xlim)
        ax.set_xlabel(r'Frequency ($\mu \rm Hz$)')
        ax.set_ylabel(r'SNR')
        ax.legend()
        return fig

def plot_trace(stage):
    """ Make a trace plot of the MCMC chains
    """

    import pymc3 as pm

    if type(stage) == pbjam.priors.kde:
        # TODO - make this work for kde
        print('Traceplot for kde not yet implimented')

    if type(stage) == pbjam.asy_peakbag.asymptotic_fit:
        # TODO - make this work for asy_peakbag
        print('Traceplot for asy_peakbag not yet implimented')

    if type(stage) == pbjam.peakbag:
        pm.traceplot(stage.samples)


# Asy_peakbag
def plot_start(self):
    """ Plots the starting model as a diagnotstic.
    """

    dnu = 10**np.median(self.start_samples, axis=0)[0]
    xlim = [min(self.f[self.sel])-dnu, max(self.f[self.sel])+dnu]
    fig, ax = plt.subplots(figsize=[16,9])
    ax.plot(self.f, self.s, 'k-', label='Data', alpha=0.2)
    smoo = dnu * 0.005 / (self.f[1] - self.f[0])
    kernel = conv.Gaussian1DKernel(stddev=smoo)
    smoothed = conv.convolve(self.s, kernel)
    ax.plot(self.f, smoothed, 'k-', label='Smoothed', lw=3, alpha=0.6)
    ax.plot(self.f[self.sel], self.model(self.start_samples.mean(axis=0)),
            'r-', label='Start model', alpha=0.7)
    ax.set_ylim([0, smoothed.max()*1.5])
    ax.set_xlim(xlim)
    ax.set_xlabel(r'Frequency ($\mu \rm Hz$)')
    ax.set_ylabel(r'SNR')
    ax.legend()
    return fig

# Peakbag
def plot_linewidth(self, thin=10):
    """ Plot estimated line width as a function of scaled n.
    """

    fig, ax = plt.subplots(1, 2, figsize=[16,9])

    if self.gp0 != []:

        n_new = np.linspace(-0.2, 1.2, 100)[:,None]
        with self.pm_model:
            f_pred0 = self.gp0.conditional("f_pred0", n_new)
            f_pred2 = self.gp2.conditional("f_pred2", n_new)
            self.pred_samples = pm.sample_posterior_predictive(self.samples,
                           vars=[f_pred0, f_pred2], samples=1000)
        plot_gp_dist(ax[0], self.pred_samples["f_pred0"], n_new)
        plot_gp_dist(ax[1], self.pred_samples["f_pred2"], n_new)

        for i in range(0, len(self.samples), thin):
            ax[0].scatter(self.n,
                          self.samples['ln_width0'][i, :], c='k', alpha=0.3)
            ax[1].scatter(self.n,
                          self.samples['ln_width2'][i, :], c='k', alpha=0.3)


    else:
        for i in range(0, len(self.samples), thin):
            ax[0].scatter(self.n,
                          np.log(self.samples['width0'][i, :]), c='k', alpha=0.3)
            ax[1].scatter(self.n,
                          np.log(self.samples['width2'][i, :]), c='k', alpha=0.3)

    ax[0].set_xlabel('normalised order')
    ax[1].set_xlabel('normalised order')
    ax[0].set_ylabel('ln line width')
    ax[1].set_ylabel('ln line width')
    ax[0].set_title('Radial modes')
    ax[1].set_title('Quadrupole modes')
    return fig

def plot_height(self, thin=10):
    """ Plots the estimated mode height.
    """

    fig, ax = plt.subplots(figsize=[16,9])
    for i in range(0, len(self.samples), thin):
        ax.scatter(self.samples['l0'][i, :], self.samples['height0'][i, :])
        ax.scatter(self.samples['l2'][i, :], self.samples['height2'][i, :])
    return fig

def plot_ladder(self, thin=10, alpha=0.2):
    """
    Plots the ladder data and models from the samples

    Parameters
    ----------
    thin: int
        Uses every other thin'th value from the samkles, i.e. [::thin].
    alpha: float64
        The alpha to use for plotting the models from samples.

    """

    n = self.ladder_s.shape[0]
    fig, ax = plt.subplots(n, figsize=[16,9])
    for i in range(n):
        for j in range(0, len(self.samples), thin):
            mod = self.model(self.samples['l0'][j],
                             self.samples['l2'][j],
                             self.samples['width0'][j],
                             self.samples['width2'][j],
                             self.samples['height0'][j],
                             self.samples['height2'][j],
                             self.samples['back'][j])
            ax[i].plot(self.ladder_f[i, :], mod[i, :], c='r', alpha=alpha)
        ax[i].plot(self.ladder_f[i, :], self.ladder_s[i, :], c='k')
        ax[i].set_xlim([self.ladder_f[i, 0], self.ladder_f[i, -1]])
    ax[n-1].set_xlabel(r'Frequency ($\mu \rm Hz$)')
    fig.tight_layout()
    return fig



""" Setup PBjam sessions and perform mode ID and peakbagging

This module contains the input layer for setting up PBjam sessions for 
peakbagging solar-like oscillators. This is the easiest way to handle targets 
in PBjam.

It's possible to manually initiate star class instances and do all the fitting
that way, but it's simpler to just use the session class, which handles
everything, including formatting of the inputs.

A jam session is started by initializing the session class instance with a
target ID, numax, a large separation, effective temperature and Gaia bp_rp 
color. 

Lists of the above can be provided for multiple targets, but it's often simpler
to just provide PBjam with a dictionary or Pandas dataframe. See mytgts.csv
for a template.

Custom timeseries or periodograms can be provided as either file pathnames,
numpy arrays, or lightkurve.LightCurve/lightkurve.periodogram objects. If
nothing is provided PBjam will download the data automatically using 
LightKurve.

Specific quarters, campgains or sectors can be requested with the relevant
keyword (i.e., 'quarter' for KIC, etc.). If none of these are provided, PBjam
will download all available data, picking the long cadence versions by default.

Once initialized, the session class contains a list of star class instances
for each requested target, with associated spectra for each.

The next step is to perform a mode ID on the spectra. At the moment PBjam
only supports use of the asymptotic relation mode ID method. Additional methods
can be added in future.

Finally the peakbagging method takes the output from the modeID and performs
a proper HMC peakbagging run to get the unparameterized mode frequencies.

Plotting the results of each stage is also possible.

Note
----

For automatic download the long cadence data set is used by default, so set
the cadence to 'short' for main-sequence targets.
"""

import lightkurve as lk
import numpy as np
import astropy.units as units
import pandas as pd
import os, glob, warnings
from .star import star


def organize_sess_dataframe(vardf):
    """ Takes input dataframe and tidies it up.

    Checks to see if required columns are present in the input dataframe,
    and adds optional columns if they don't exists, containing None values.

    Parameters
    ----------
    vardf : Pandas.DataFrame
        Input dataframe
        
    """
    
    keys = ['ID', 'numax', 'dnu', 'numax_err', 'dnu_err']
    if not any(x not in keys for x in vardf.keys()):
        raise(KeyError, 'Some of the required keywords were missing.')

    N = len(vardf)
    singles = ['cadence', 'campaign', 'sector', 'month', 'quarter', 'mission']
    doubles = ['teff', 'bp_rp']

    for key in singles:
        if key not in vardf.keys():
            vardf[key] = np.array([None]*N)

    for key in doubles:
        if key not in vardf.keys():
            vardf[key] = np.array([None]*N)
            vardf[key+'_err'] = np.array([None]*N)

    if 'timeseries' not in vardf.keys():
        format_col(vardf, None, 'timeseries')
    if 'psd' not in vardf.keys():
        format_col(vardf, None, 'psd')


def organize_sess_input(**vardct):
    """ Takes input and organizes them in a dataframe.

    Checks to see if required inputs are present and inserts them into a
    dataframe. Any optional columns that are not included in the input are
    added as None columns.

    Parameters
    ----------
    vardct : objects
        Variable inputs to Session class to be arranged into a dataframe

    Returns
    -------
    vardf : Pandas.DataFrame
        Dataframe containing the inputs from Session class call.
        
    """
    
    vardf = pd.DataFrame({'ID': np.array(vardct['ID']).reshape((-1, 1)).flatten()})

    N = len(vardf)
    singles = ['cadence', 'campaign', 'sector', 'month', 'quarter', 'mission']
    doubles = ['numax', 'dnu', 'teff', 'bp_rp']

    for key in singles:
        if not vardct[key]:
            vardf[key] = np.array([None]*N)
        else:
            vardf[key] = vardct[key]

    for key in doubles:
        if not vardct[key]:
            vardf[key] = np.array([None]*N)
            vardf[key+'_err'] = np.array([None]*N)
        else:
            vardf[key] = np.array(vardct[key]).reshape((-1, 2))[:, 0].flatten()
            vardf[key+'_err'] = np.array(vardct[key]).reshape((-1, 2))[:, 1].flatten()
    return vardf


def launch_query(id, cache, lkwargs):
    """ Search for target on MAST server.

    Get all the lightcurves available for a target id, using options in kwargs
    dictionary. The lightcurves are downloaded using the lightkurve API, and
    the target ID must therefore be parseable by lightkurve.

    Parameters
    ----------
    id : string
        Target id, must be resolvable by Lightkurve.

    cache : str
        Directory to download the lightcurves into.

    lkwargs : dictionary containing keywords for the LightKurve search.
        cadence, quarter, campaign, sector, month.

    Returns
    -------
    search_results : list
        List of fits files for the requested target
        
    """

    print(f'Querying MAST for {id}')
    search_results = lk.search_lightcurvefile(target=id, **lkwargs)
    if len(search_results) == 0:
        warnings.warn('LightKurve did not return %s cadence data for %s' % (lkwargs['cadence'], id))
        return []
    else:
        return search_results.download_all(download_dir=cache)


def sort_lc(lc):
    """ Sort a lightcurve in Lightkurve object.

    Lightkurve lightcurves are not necessarily sorted in time, which causes
    an error in periodogram.

    Parameters
    ----------
    lc : Lightkurve.LightCurve instance
        Lightkurve object to be modified

    Returns
    -------
    lc : Lightkurve.LightCurve instance
        The sorted Lightkurve object
        
    """

    sidx = np.argsort(lc.time)
    lc.time = lc.time[sidx]
    lc.flux = lc.flux[sidx]
    return lc


def clean_lc(lc):
    """ Perform Lightkurve operations on object.

    Performes basic cleaning of a light curve, removing nans, outliers,
    median filtering etc.

    Parameters
    ----------
    lc : Lightkurve.LightCurve instance
        Lightkurve object to be cleaned

    Returns
    -------
    lc : Lightkurve.LightCurve instance
        The cleaned Lightkurve object
        
    """

    lc = lc.remove_nans().normalize().flatten().remove_outliers()
    return lc

def set_cadence(lkwargs):
    """ Select the cadence 
    
    Determines the extension to use later in the lookup of cached fits files,
    to be passed to LightKurve for online lookup.
    
    If no cadence argument is passed it will default to long cadence. 
    
    Parameters
    ----------
    lkwargs : dict
        Dictionary to be passed to LightKurve
        
    Returns
    -------
    ext : str
        Fits file short/long cadence extension
        
    """
    if not lkwargs['cadence']:
        lkwargs['cadence'] = 'long'
    
    if lkwargs['cadence'] == 'short':
        ext = '*_slc.fits'
    elif lkwargs['cadence'] == 'long':
        ext = '*_lc.fits'
    else:
        raise TypeError('Unrecognized cadence input for %s' % (id))
    return ext

def set_cache_dir(download_dir):
    """ Determine which directory to use as cache
    
    Parameters
    ----------
    download_dir : str
        None or path to store results from a star
    
    Returns
    -------
    cache_dir : str
        Path to store results from a star
        
    """
    
    if not download_dir:
        cache_dir = os.path.join(*[os.path.expanduser('~'), 
                                     '.lightkurve-cache'])
    else:
        cache_dir = download_dir
    
    return cache_dir    

def set_mission(ID, lkwargs):
    """ Set mission keyword in lkwargs
    
    If no mission is selected will attempt to figure it out based on any
    prefixes in the ID string, and add this to the LightKurve keywords 
    arguments dictionary.
    
    Parameters
    ----------
    ID : str
        ID string of the target
    lkwargs : dict
        Dictionary to be passed to LightKurve
        
    """
    if lkwargs['mission'] is None:
        if ('kic' in ID.lower()) or ('kplr' in ID.lower()) :
            lkwargs['mission'] = 'kepler'
        elif ('ktwo' in ID.lower()) or ('epic' in ID.lower()):
            lkwargs['mission'] = 'k2'
        elif ('tic' in ID.lower()) or ('tess' in ID.lower()):
            lkwargs['mission'] = 'tess'
        else:
            warnings.warn('Unknown mission selected. MAST might not understand.')

def lookup_cached_files(id, cache_dir, ext):
    """ Look through the local cache directory for target files
    
    Looks through the local cache directory for any files matching the ID of 
    the target, and with the requested extension.
    
    Parameters
    ----------
    id : str
        Input ID for the target
    cache_dir : str
        Path to the cache directory
    ext : str
        Fits file short/long cadence extension
    
    return : list
        List of file names matching the search criteria
        
    """
    
    if isinstance(id, str):
        baseid = id.lower()
        for prefix in ['kic','epic','tic','kplr']:
            baseid = baseid.replace(prefix, '')
        baseid = str(int(baseid))
      
    tgtfiles = glob.glob(os.path.join(*[cache_dir, 'mastDownload', '*', 
                                        f'*{baseid}*', ext]))
    return tgtfiles

def query_lightkurve(id, download_dir, use_cached, lkwargs):
    """ Check cache for fits file, or download it.

    Based on use_cached flag, will look in the cache for fits file
    corresponding to request id star. If nothing is found in cached it will be
    downloaded from the MAST server.

    Parameters
    ----------
    id : string
        Identifier for the requested star. Must be resolvable by Lightkurve
    lkwargs : dict
        Dictionary containing keywords for the Lightkurve search.
        cadence, quarter, campaign, sector, month.
    use_cached: bool
        Whether or not to used data in the Lightkurve cache.

    Note:
    -----
    Prioritizes long cadence over short cadence unless otherwise specified.

    """
    
    cache_dir = set_cache_dir(download_dir)
    
    set_mission(id, lkwargs)
    
    ext = set_cadence(lkwargs)

    tgtfiles = lookup_cached_files(id, cache_dir, ext)

    if (use_cached and (len(tgtfiles) != 0)):
        lc_col = [lk.open(n) for n in tgtfiles]
        
    elif (not use_cached) or (use_cached and (len(tgtfiles) == 0)):
        if (use_cached and (len(tgtfiles) == 0)):
            warnings.warn('Could not find %s cadence data for %s in cache, checking MAST...' % (lkwargs['cadence'], id))

        lc_col = launch_query(id, cache_dir, lkwargs)

        if len(lc_col) == 0:
            raise ValueError("Could not find %s cadence data for %s in cache or on MAST" % (lkwargs['cadence'], id))
    else:
        raise ValueError('Could not find any cached data, and failed to access MAST')
    
    # Perform reduction on first lc of the lc collection and append the rest
    lc0 = clean_lc(lc_col[0].PDCSAP_FLUX)
    for i, lc in enumerate(lc_col[1:]):
        lc0 = lc0.append(clean_lc(lc.PDCSAP_FLUX))
        
    return lc0

      
def arr_to_lk(x, y, name, typ):
    """ LightKurve object from input.

    Creates either a lightkurve.LightCurve or lightkurve.periodogram object
    from the input arrays.

    Parameters
    ----------
    x : list-like
        First column of timeseries or periodogram (time/frequency).
    y : list-like
        Second column of timeseries or periodogram (flux/power).
    name : string
        Target ID
    typ : string
        Either timeseries or periodogram.

    Returns
    -------
    lkobj : object
        Either lightkurve.LightCurve or lightkurve.periodogram object
        depending on typ.

    """
    
    if typ == 'timeseries':
        return lk.LightCurve(time=x, flux=y, targetid=name)
    elif typ == 'psd':
        return lk.periodogram.Periodogram(x*units.microhertz,
                                          units.Quantity(y, None),
                                          targetid=name)
    else:
        raise KeyError("Don't modify anything but psd and timeseries cols")


def format_col(vardf, col, key):
    """ Add timeseries or psd column to dataframe based on input

    Based on the contents of col, will try to format col and add it as a column
    to vardf with column name key. col can be many things, so the decision is
    based mainly on the dimensionality of col.

    If dim = 0, it's assumed that col is either None, or a string, (for the
    latter it assumes there is then only one target).

    If dim = 1, it's assumed that col is a list-like object, consisting of
    either None or strings, these are passed along without modification.

    If dim = 2, col is assumed to be either a time series or power spectrum
    of shape (2,M), with time/frequency in 1st row and flux/power in the
    second.

    If dim = 3, it is assumed to be list of (2,M) arrays.

    In both of the latter cases col is converted to Lightkurve object(s).

    Parameters
    ----------
    vardf : pandas.DataFrame instance
        Pandas Dataframe instance to which either a timeseries or psd column
        will be added.
    col : object
        Input from Session call, corresponding to key
    key : str
        Name of column to add to vardf

    """
    
    N = np.shape(vardf['ID'])[0]

    col = np.array(col, dtype=object)

    # If dim = 0, it's either none or a string
    if col.ndim == 0:
        if not col:
            # If none, then multiply up to length of ID
            vardf[key] = np.array([None]*N)
        else:
            # If string, then single target
            vardf[key] = np.array(col).reshape((-1, 1)).flatten()

    # If dim = 1, it's either a list of nones, strings or lightkurve objects
    elif col.ndim == 1:
        vardf[key] = col

    # if dim = 2, it's an array or tuple, with time and flux
    elif col.ndim == 2:
        x = np.array(col[0, :], dtype=float)
        y = np.array(col[1, :], dtype=float)
        vardf[key] = [arr_to_lk(x, y, vardf['ID'][0], key)]

    # If dim = 3, it's a list of arrays or tuples
    elif col.ndim == 3:
        temp = np.array([], dtype=object)
        for i in range(N):
            x = np.array(col[i, 0, :], dtype=float)
            y = np.array(col[i, 1, :], dtype=float)
            temp = np.append(temp,
                             np.array([arr_to_lk(x, y, vardf.loc[i, 'ID'], key)]))
        vardf[key] = temp
    else:
        print('Unhandled exception')


def lc_to_lk(vardf, download_dir, use_cached=True):
    """ Convert time series column in dataframe to lk.LightCurve object

    Goes through the timeseries column in the dataframe and tries to convert
    it to a Lightkurve.LightCurve object. If string, it's assumed to be a file
    path name, if None, it will query the LightCurve cache locally or MAST if
    nothing is found. Skips column entries which are already LightKurve objects
    or if a psd for the star in question exists in the psd column.

    Parameters
    ----------
    vardf : pandas.DataFrame instance
        Dataframe containing a 'timeseries' column consisting of None, strings
        or Lightkurve.LightCurve objects.

    """

    tinyoffset = 1  # to avoid cases LC median = 0 (lk doesn't like it)
    key = 'timeseries'
    for i, id in enumerate(vardf['ID']):
        if isinstance(vardf.loc[i, key], str):
            t, d = np.genfromtxt(vardf.loc[i, key], usecols=(0, 1)).T
            d += tinyoffset
            vardf.at[i, key] = arr_to_lk(t, d, vardf.loc[i, 'ID'], key)
        elif not vardf.loc[i, key]:
            if vardf.loc[i, 'psd']:
                pass
            else:
                D = {x: vardf.loc[i, x] for x in ['cadence', 'month', 'sector',
                                                  'campaign', 'quarter', 
                                                  'mission']}
                lk_lc = query_lightkurve(id, download_dir, use_cached, D)
                vardf.at[i, key] = lk_lc

        elif vardf.loc[i, key].__module__ == lk.lightcurve.__name__:
            pass
        else:
            raise TypeError("Can't handle this type of time series object")

        if vardf.loc[i, key]:
            sort_lc(vardf.loc[i, key])


def lk_to_pg(vardf):
    """ Convert psd column in dataframe to Lightkurve periodgram object list

    Takes whatever is in the psd column of a dataframe and tries to turn it
    into a Lightkurve.periodogram object. If column entry is a string, it
    assumes it's a path name, if None, it will try to take what's in the
    timeseries column and compute periodogram based on that. If a periodogram
    object is already present, it'll just flatten it (to be safe) and then
    continue on.

    Parameters
    ----------
    vardf : pandas.DataFrame instance
        Dataframe containing a 'psd' column consisting of None, strings or
        Lightkurve.periodogram objects.

    """

    key = 'psd'
    for i, id in enumerate(vardf['ID']):
        if isinstance(vardf.loc[i, key], str):
            f, s = np.genfromtxt(vardf.loc[i, key], usecols=(0, 1)).T
            vardf.at[i, key] = arr_to_lk(f, s, vardf.loc[i, 'ID'], key)
        elif not vardf.loc[i, key]:
            lk_lc = vardf.loc[i, 'timeseries']
            vardf.at[i, key] = lk_lc.to_periodogram(freq_unit=units.microHertz, normalization='psd').flatten()

        elif vardf.loc[i, key].__module__ == lk.periodogram.__name__:
            vardf.at[i, key] = vardf.loc[i, key].flatten()
        else:
            raise TypeError("Can't handle this type of time series object")


class session():
    """ Main class used to initiate peakbagging.

    Use this class to initialize a star class instance for one or more targets.

    Once initialized, calling this class instance will execute a complete
    peakbagging run.

    Data can be provided in multiple different ways, the simplest of which is
    just to let PBjam query the MAST server. Otherwise arrays of
    timeseries/power spectra, lightkurve.LightCurve/lightkurve.periodogram,
    or just path names as strings, is also possible.

    The physical parameters, numax, dnu, teff, must each be provided at least
    as a list of length 2 for each star. This should contain the parameter
    value and it's error.

    For multiple targets all the above can be provided as lists, but the
    easiest way is to simply provide a dataframe from a csv file.

    Examples
    --------
    Peakbagging run for a single target:

    >>> jam_sess = pbjam.session(ID =  '4448777',  numax = [220.0, 3.0],
                             dnu = [16.97, 0.01], teff = [4750, 100],
                             bp_rp = [1.34, 0.01], cadence = 'short')
    >>> jam_sess()

    Peakbagging run for multiple targets:
    >>> jam_sess = pbjam.session(dictlike = mydataframe)
    >>> jam_sess()

    By default, PBjam will download all the available data, favoring long
    cadence. Cadence and specific observing seasons (quarter, month, campagin,
    sector) can be specified for more detailed control. The data is downloaded
    using the Lightkurve API, and cleaned on the timeseries level by removing
    NaNs, and outliers, and flattening the lightcurve (again using LightKurve's
    API). Note that the flattening process may impact the appearance of the
    granulation background.

    Parameters
    ----------
    ID : string, int
        Target identifier, if custom timeseries/periodogram is provided, it
        must be resolvable by LightKurve (KIC, TIC, EPIC, HD, etc.)

    numax : list
        List of the form [numax, numax_error], list of lists for multiple
        targets

    dnu : list
        List of the form [dnu, dnu_error], list of lists for multiple targets

    teff : list, optional
        List of the form [teff, teff_error], list of lists for multiple targets

    bp_rp : list, optional
        List of the form [bp_rp, bp_rp_error], list of lists for multiple
        targets

    timeseries : object, optional
        Timeseries input. Leave as None for PBjam to download it automatically.
        Otherwise, arrays of shape (2,N), lightkurve.LightCurve objects, or
        strings for pathnames are accepted.

    psd : object, optional
        Periodogram input. Leave as None for PBjam to use Timeseries to compute
        it for you. Otherwise, arrays of shape (2,N), lightkurve.periodogram
        objects, or strings for pathnames are accepted.

    dictlike : pandas.DataFrame or dictionary, optional
        DataFrame, dictionary, record array with a list of targets, and their
        properties. If string, PBjam will assume it's a pathname to a csv file.
        Specify timeseries and psd columns with file pathnames to use manually
        reduced data.

    store_chains : bool, optional
        Flag for storing all the full set of samples from the MCMC run.
        Warning, if running multiple targets, make sure you have enough memory.

    nthreads : int, optional
        Number of multiprocessing threads to use to perform the fit. For long
        cadence data 1 is best, more will just add parallelization overhead.
        Untested on short cadence.

    use_cached : bool, optional
        Flag for using cached data. If fitting the same targets multiple times,
        use to this to not download the data every time.

    cadence : string, optional
        Argument for lightkurve to download correct data type. Can be 'short'
        or 'long'. 'long' is default setting, so if you're looking at main
        sequence stars, make sure to manually set 'short'.
    
    campaign : int, optional
        Argument for lightkurve when requesting K2 data.

    sector : int, optional
        Argument for lightkurve when requesting TESS data.

    month : int, optional
        Argument for lightkurve when requesting Kepler short cadence data.

    quarter : int, optional
        Argument for lightkurve when requesting Kepler data.
    
    mission : str, optional
        Which mission to use data from. Default is all, so if your target has
        been observed by, e.g., both Kepler and TESS, LightKurve will throw
        and error.

    make_plots : bool, optional
        Whether or not to automatically generate diagnostic plots for the 
        different stages of the peakbagging process
    
    path : str, optional
        Path to store the plots and results for the various stages of the 
        peakbagging process
    
    download_dir : str, optional
        Directory to cache lightkurve downloads. Lightkurve will place the fits
        files in the default lightkurve cache path in your home directory. 
    
    model_type : str, optional
        Argument passed to peakbag, defines which model type to be used to 
        represent the mode linewidths. Options are 'simple' or 'model_gp'.
        
    
    Attributes
    ----------
    nthreads : int
        Number of multiprocessing threads to use to perform the HMC peakbag 
        fit. 

    store_chains : bool
        Flag for storing all the full set of samples from the MCMC run.
        Warning, if running multiple targets, make sure you have enough memory.

    stars : list
        Session will store star class instances in this list, based on the
        requested targets.

    pb_model_type : str
        Argument passed to peakbag, defines which model type to be used to 
        represent the mode linewidths. Options are 'simple' or 'model_gp'.
        
    """

    def __init__(self, ID=None, numax=None, dnu=None, teff=None, bp_rp=None,
                 timeseries=None, psd=None, dictlike=None, use_cached=False, 
                 cadence=None, campaign=None, sector=None, month=None, 
                 quarter=None, mission=None, path=None, download_dir=None):

        self.stars = []
        
        if isinstance(dictlike, (dict, np.recarray, pd.DataFrame, str)):
            if isinstance(dictlike, str):
                vardf = pd.read_csv(dictlike)
            else:
                try:
                    vardf = pd.DataFrame.from_records(dictlike)
                except TypeError:
                    print('Unrecognized type in dictlike. Must be able to convert to dataframe through pandas.DataFrame.from_records()')

            if any([ID, numax, dnu, teff, bp_rp]):
                warnings.warn('Dictlike provided as input, ignoring other input fit parameters.')

            organize_sess_dataframe(vardf)

        elif ID:
            vardf = organize_sess_input(ID=ID, numax=numax, dnu=dnu, teff=teff,
                                        bp_rp=bp_rp, cadence=cadence,
                                        campaign=campaign, sector=sector,
                                        month=month, quarter=quarter, 
                                        mission=mission)
            format_col(vardf, timeseries, 'timeseries')
            format_col(vardf, psd, 'psd')

        # Take whatever is in the timeseries column of vardf and make it an
        # lk.lightcurve object or None

        lc_to_lk(vardf, download_dir, use_cached=use_cached)
        
        # Take whatever is in the timeseries column of vardf and turn it into
        # a periodogram object in the periodogram column.

        lk_to_pg(vardf)


        for i in range(len(vardf)):
            self.stars.append(star(ID=vardf.loc[i, 'ID'],
                                   pg=vardf.loc[i, 'psd'],
                                   numax=vardf.loc[i, ['numax', 'numax_err']].values,
                                   dnu=vardf.loc[i, ['dnu', 'dnu_err']].values,
                                   teff=vardf.loc[i, ['teff', 'teff_err']].values,
                                   bp_rp=vardf.loc[i, ['bp_rp', 'bp_rp_err']].values,
                                   path=path))

        for i, st in enumerate(self.stars):
            if st.numax[0] > st.f[-1]:
                warnings.warn("Input numax is greater than Nyquist frequeny for %s" % (st.ID))

    def __call__(self, bw_fac=1, tune=1500, norders=8, model_type='simple', 
                 verbose=False, nthreads=1, store_chains=False, 
                 make_plots=False):
        """ The doitall script

        Calling session will by default do asymptotic mode ID and peakbagging
        for all stars in the session.

        Parameters
        ----------
        norders : int
            Number of orders to include in the fits
            
        """
        
        self.pb_model_type = model_type

        for i, st in enumerate(self.stars):
            try:
                st(bw_fac=bw_fac, tune=tune, norders=norders, 
                   model_type=self.pb_model_type, verbose=verbose, 
                   make_plots=make_plots, store_chains=store_chains, 
                   nthreads=nthreads)
                
                self.stars[i] = None
                    
            except Exception as ex:
                 message = "Star {0} produced an exception of type {1} occurred. Arguments:\n{2!r}".format(st.ID, type(ex).__name__, ex.args)
                 print(message)
            



import setuptools

__version__ = []
exec(open('pbjam/version.py').read())

setuptools.setup(
    name="pbjam",
    version=__version__,
    author="Martin Nielsen, Guy Davies, Oliver Hall",
    author_email="m.b.nielsen.1@bham.ac.uk",
    description="A package for peakbagging solar-like oscillators",
    long_description=open("README.rst").read(),
    url="https://pbjam.readthedocs.io/",
    packages=['pbjam'],
    install_requires=open('requirements.txt').read().splitlines(),
    extras_require={'docs': ["nbsphinx"]},
    include_package_data=True,
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Development Status :: 3 - Alpha"
    ],
)



#import matplotlib
#matplotlib.use('Agg')

import numpy as np
import pymc3 as pm
import warnings
from .plotting import plotting

class peakbag(plotting):
    """
    Class for PBjam peakbagging.

    This class allows for simple manipulation of the data, the fitting of a
    PyMC3 model to the data, and some plotting of results functionality.

    Examples
    --------
    Star API interaction (recommended)

    >>> from pbjam import star
    >>> star = pb.star(kic, pg, numax, dnu, teff, bp_rp, store_chains=True, nthreads=4)
    >>> star.run_epsilon()
    >>> star.run_asy_peakbag(norders=7)
    >>> star.run_peakbag()

    Lower level API interaction.

    >>> import PBjam
    >>> pbag = pbjam.peakbag(frequency, snr, asy_result)
    >>> pbag.sample(model_type='simple', cores=4)
    >>> pbag.plot_fit()

    Parameters
    ----------
    f : float, array
        Array of frequency bins of the spectrum (muHz). Truncated to the range
        around numax.
    snr : float, array
        Array of SNR values for the frequency bins in f (dimensionless).
    asy_result : asy_result
        The result from the asy_peakbag method.
    init : bool
        If true runs make_start and trim_ladder to prepare starting
        guesses for sampling and transforms the data onto a ladder.

    Attributes
    ----------
    f : float, ndarray
        Array of frequency bins of the spectrum (muHz). Truncated to the range
        around numax.
    snr : float, ndarray
        Array of SNR values for the frequency bins in f (dimensionless).
    asy_result : asy_result
        The result from the asy_peakbag method.
        This is a dictionary of 'modeID' and 'summary'.
        'modeID' is a DataFrame with a list of modes and basic properties.
        'summary' are summary statistics from the asymptotic_fit.
        See asy_peakbag asymptotic_fit for more details.

    """

    def __init__(self, starinst, asyinst, init=True, path=None,
                       verbose=False):

        self.pg = starinst.pg
        self.f = starinst.f
        self.s = starinst.s
        self.asy_result = asyinst
        self.norders = asyinst.norders
        if init:
            self.make_start()
            self.trim_ladder(verbose=verbose)
        self.gp0 = [] # Used for gp linewidth info.

        starinst.peakbag = self


    def make_start(self):
        """
        Function uses the information in self.asy_result (the result of the
        asymptotic peakbagging) and builds a dictionary of starting values
        for the peakbagging methods.

        """

        idxl0 = self.asy_result.modeID.ell == 0
        idxl2 = self.asy_result.modeID.ell == 2

        l0 = self.asy_result.modeID.loc[idxl0, 'nu_med'].values.flatten()
        l2 = self.asy_result.modeID.loc[idxl2, 'nu_med'].values.flatten()

        l0, l2 = self.remove_outsiders(l0, l2)

        width = 10**(np.ones(len(l0)) * self.asy_result.summary.loc['mode_width', 'mean']).flatten()
        height =  (10**self.asy_result.summary.loc['env_height', 'mean'] * \
                 np.exp(-0.5 * (l0 - 10**self.asy_result.summary.loc['numax', 'mean'])**2 /
                 (10**self.asy_result.summary.loc['env_width', 'mean'])**2)).flatten()
        back = np.ones(len(l0))

        self.parnames = ['l0', 'l2', 'width0', 'width2', 'height0', 'height2',
                         'back']

        pars = [l0, l2, width, width, height, 0.7*height, back]

        self.start ={x:y for x,y in zip(self.parnames, pars)}
#        self.start = {'l0': l0, 'l2': l2, 'width0': width, 'width2': width,
#                      'height0': height, 'height2': height*0.7,
#                      'back': np.ones(len(l0))}
        self.n = np.linspace(0.0, 1.0, len(self.start['l0']))[:, None]

    def remove_outsiders(self, l0, l2):
        """ Drop outliers

        Drops modes where the guess frequency is outside of the supplied
        frequency range.

        """

        sel = np.where(np.logical_and(l0 < self.f.max(), l0 > self.f.min()))
        return l0[sel], l2[sel]

    def trim_ladder(self, lw_fac=10, extra=0.01, verbose=False):
        """
        This function makes ladders and then selects only the orders
        in the ladders that have modes that are to be fitted,
        i.e, it trims the ladder.

        Each ladder rung is constructed so that the central frequency is
        the mid point between the l=0 and l=2 modes as determined by the
        info in the asy_result dictionary.

        The width of the rung is determined by d02, line width, and some
        parameters lw_fac and extra.  The basic calculation is the width is:
        d02 + (lw_fac * line_width) + (extra * dnu)

        Parameters
        ----------
        lw_fac: float
            The factor by which the mode line width is multiplied in order
            to contribute to the rung width.
        extra: float
            The factor by which dnu is multiplied in order to contribute to
            the rung width.

        """

        d02 = 10**self.asy_result.summary.loc['d02','mean']
        d02_lw = d02 + lw_fac * 10**self.asy_result.summary.loc['mode_width', 'mean']
        w = d02_lw + (extra * 10**self.asy_result.summary.loc['dnu', 'mean'])
        bw = self.f[1] - self.f[0]
        w /= bw
        if verbose:
            print(f'w = {int(w)}')
            print(f'bw = {bw}')
        ladder_trim_f = np.zeros([len(self.start['l0']), int(w)])
        ladder_trim_s = np.zeros([len(self.start['l0']), int(w)])
        for idx, freq in enumerate(self.start['l0']):
            loc_mid_02 = np.argmin(np.abs(self.f - (freq - d02/2.0)))
            if loc_mid_02 == 0:
                warnings.warn('Did not find optimal rung location')
            if verbose:
                print(f'loc_mid_02 = {loc_mid_02}')
                print(f'w/2 = {int(w/2)}')
            ladder_trim_f[idx, :] = \
                self.f[loc_mid_02 - int(w/2): loc_mid_02 - int(w/2) + int(w)]
            ladder_trim_s[idx, :] = \
                self.s[loc_mid_02 - int(w/2): loc_mid_02 - int(w/2) + int(w) ]
        self.ladder_f = ladder_trim_f
        self.ladder_s = ladder_trim_s

    def lor(self, freq, w, h):
        """
        This function calculates a lorentzian for each rung of the frequency
        ladder.  The ladder is a 2D array.  freq, w, and h should be 1D arrays
        of length that matches the height of the ladder.  No checks are made
        for this so as to reduce overheads.

         Parameters
         ----------
         freq : float, ndarray
            A length H array of Lorentzian central frequencies where H is the
            height of self.ladder_f .
         w : float, ndarray
            A length H array of Lorentzian width where H is the
            height of self.ladder_f .
         h : float, ndarray
            A length H array of Lorentzian heightswhere H is the
            height of self.ladder_f .

        Returns
        -------
        lorentzians : float, ladder
           A ladder containing one Lorentzian per rung.

        """

        norm = 1.0 + 4.0 / w**2 * (self.ladder_f.T - freq)**2
        return h / norm

    def model(self, l0, l2, width0, width2, height0, height2, back):
        """
        Calcuates a simple model of a flat backgroud plus two lorentzians
        for each rung of self.ladder_f .

        Parameters
        ----------
        l0 : float, ndarray
            A length H array of l=0 mode central frequencies where H is the
            height of self.ladder_f .
        l2 : float, ndarray
            A length H array of l=2 mode central frequencies where H is the
            height of self.ladder_f .
        width0 : float, ndarray
            A length H array of l=0 mode widths where H is the
            height of self.ladder_f .
        width2 : float, ndarray
            A length H array of l=2 mode widths where H is the
            height of self.ladder_f .
        height0 : float, ndarray
            A length H array of l=0 mode heights where H is the
            height of self.ladder_f .
        height2 : float, ndarray
            A length H array of l=2 mode heights where H is the
            height of self.ladder_f .
        back : float, ndarray
            A length H array of background values where H is the
            height of self.ladder_f .

        Returns
        -------
        mod : float, ndarray
            A 2D array (or 'ladder') containing the calculated model.

        """

        mod = np.ones(self.ladder_f.shape).T * back
        mod += self.lor(l0, width0, height0)
        mod += self.lor(l2, width2, height2)
        return mod.T

    def init_model(self, model_type):
        """
        TODO - Need to describe what is happening here.
        Complete docs when model is settled on.  Probably quiet a
        long docs needed to explain.

        """

        dnu = 10**self.asy_result.summary.loc['dnu', 'mean']
        dnu_fac = 0.03 # Prior on mode frequency has width 3% of Dnu.
        height_fac = 0.4 # Lognorrmal prior on height has std=0.4.
        width_fac = 1.0 # Lognorrmal prior on width has std=1.0.
        back_fac = 0.5 # Lognorrmal prior on back has std=0.5.
        N = len(self.start['l2'])

        with self.pm_model:

            if model_type != 'model_gp':
                if model_type != 'simple': # defaults to simple if bad input
                    warnings.warn('Model not defined - using simple model')
                width0 = pm.Lognormal('width0', mu=np.log(self.start['width0']),
                                  sigma=width_fac, shape=N)
                width2 = pm.Lognormal('width2', mu=np.log(self.start['width2']),
                                  sigma=width_fac, shape=N)

                self.init_sampler = 'adapt_diag'
                self.target_accept = 0.9

            elif model_type == 'model_gp':
                warnings.warn('This model is developmental - use carefully')
                # Place a GP over the l=0 mode widths ...
                m0 = pm.Normal('gradient0', 0, 10)
                c0 = pm.Normal('intercept0', 0, 10)
                sigma0 = pm.Lognormal('sigma0', np.log(1.0), 1.0)
                ls = pm.Lognormal('ls', np.log(0.3), 1.0)
                mean_func0 = pm.gp.mean.Linear(coeffs=m0, intercept=c0)
                cov_func0 = sigma0 * pm.gp.cov.ExpQuad(1, ls=ls)
                self.gp0 = pm.gp.Latent(cov_func=cov_func0, mean_func=mean_func0)
                ln_width0 = self.gp0.prior('ln_width0', X=self.n)
                width0 = pm.Deterministic('width0', pm.math.exp(ln_width0))
                # and on the l=2 mode widths
                m2 = pm.Normal('gradient2', 0, 10)
                c2 = pm.Normal('intercept2', 0, 10)
                sigma2 = pm.Lognormal('sigma2', np.log(1.0), 1.0)
                mean_func2 = pm.gp.mean.Linear(coeffs=m2, intercept=c2)
                cov_func2 = sigma2 * pm.gp.cov.ExpQuad(1, ls=ls)
                self.gp2 = pm.gp.Latent(cov_func=cov_func2, mean_func=mean_func2)
                ln_width2 = self.gp2.prior('ln_width2', X=self.n)
                width2 = pm.Deterministic('width2', pm.math.exp(ln_width2))

                self.init_sampler = 'advi+adapt_diag'
                self.target_accept = 0.99


            l0 = pm.Normal('l0', self.start['l0'], dnu*dnu_fac, shape=N)

            l2 = pm.Normal('l2', self.start['l2'], dnu*dnu_fac, shape=N)

            height0 = pm.Lognormal('height0', mu=np.log(self.start['height0']),
                                    sigma=height_fac, shape=N)
            height2 = pm.Lognormal('height2', mu=np.log(self.start['height2']),
                                    sigma=height_fac, shape=N)
            back = pm.Lognormal('back', mu=np.log(1.0), sigma=back_fac, shape=N)

            limit = self.model(l0, l2, width0, width2, height0, height2, back)
            pm.Gamma('yobs', alpha=1, beta=1.0/limit, observed=self.ladder_s)



    def __call__(self, model_type='simple', tune=1500, nthreads=1, maxiter=4,
                     advi=False):
        """
        Function to perform the sampling of a defined model.

        Parameters
        ----------
        model_type : str
            Defaults to 'simple'.
            Can be either 'simple' or 'model_gp' which sets the type of model
            to be fitted to the data.
        tune : int
            Numer of tuning steps passed to pym3.sample
        nthreads : int
            Number of cores to use - passed to pym3.sample
        maxiter : int
            Number of times to attempt to reach convergence
        advo : bool
            Whether or not to fit using the fullrank_advi option in pymc3

        """

        self.pm_model = pm.Model()

        self.init_model(model_type=model_type)

        if advi:
            with self.pm_model:
                cb = pm.callbacks.CheckParametersConvergence(every=1000,
                                                             diff='absolute',
                                                             tolerance=0.01)

                mean_field = pm.fit(n=200000, method='fullrank_advi',
                                    start=self.start,
                                    callbacks=[cb])
                self.samples = mean_field.sample(1000)
        else:
            Rhat_max = 10
            niter = 1
            while Rhat_max > 1.05:
                if niter > maxiter:
                    warnings.warn('Did not converge!')
                    break
                with self.pm_model:
                    self.samples = pm.sample(tune=tune * niter, cores=nthreads,
                                             start=self.start,
                                             init=self.init_sampler,
                                             target_accept=self.target_accept,
                                             progressbar=False)
                Rhat_max = np.max([v.max() for k, v in pm.diagnostics.gelman_rubin(self.samples).items()])
                niter += 1

        self.summary = pm.summary(self.samples)
        self.par_names = self.summary.index


#    def simple(self):
#        """
#        Creates a simple peakbagging model in PyMC3's self.pm_model which is
#        an instance of pm.Model().
#
#        The simple model has three parameters per mode (freq, w, h) and one
#        back parameter per rung of the frequency ladder.
#
#        All parameters are independent.  For a model with additional constraints
#        see model_gp.
#
#        Priors on parameters are defined as follows:
#            lX ~ Normal(log(start[lX]), dnu*0.1)
#            widthX ~ Lognormal(log(start[widthX]), 1.0)
#            heightX ~ Lognormal(log(start[height0]), 0.4)
#            back ~ Lognormal(log(1), 0.5)
#
#        The equivalent likelihood function of the observed data is dealt with
#        using a Gamma distribution where alpha=1 and beta=1/limit where limit
#        is the model of the spectrum proposed.  Using this gamma distirbution
#        is the equivalent of stating that observed/model is distributed as
#        chi squared two degrees of freedom (see Anderson 1990 for more details).
#        The use of the Gamma distribution is much more in line with the ethos of
#        probabistic programming languages.
#        """
#
#        dnu = 10**self.asy_result.summary.loc['dnu', 'mean']
#        dnu_fac = 0.03 # Prior on mode frequency has width 3% of Dnu.
#        width_fac = 1.0 # Lognorrmal prior on width has std=1.0.
#        height_fac = 0.4 # Lognorrmal prior on height has std=0.4.
#        back_fac = 0.4 # Lognorrmal prior on back has std=0.4.
#        N = len(self.start['l2'])
#
#        with self.pm_model:
#
#            width0 = pm.Lognormal('width0', mu=np.log(self.start['width0']),
#                                  sigma=width_fac, shape=N)
#            width2 = pm.Lognormal('width2', mu=np.log(self.start['width2']),
#                                  sigma=width_fac, shape=N)
#
#            l0 = pm.Normal('l0', self.start['l0'], dnu*dnu_fac, shape=N)
#
#            l2 = pm.Normal('l2', self.start['l2'], dnu*dnu_fac, shape=N)
#
#            height0 = pm.Lognormal('height0', mu=np.log(self.start['height0']),
#                                    sigma=height_fac, shape=N)
#            height2 = pm.Lognormal('height2', mu=np.log(self.start['height2']),
#                                    sigma=height_fac, shape=N)
#            back = pm.Lognormal('back', mu=np.log(1.0), sigma=back_fac, shape=N)
#
#            limit = self.model(l0, l2, width0, width2, height0, height2, back)
#
#            pm.Gamma('yobs', alpha=1, beta=1.0/limit, observed=self.ladder_s)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))

from .version import __version__
from .priors import kde
from .session import session
from .asy_peakbag import asymp_spec_model, asymptotic_fit
from .peakbag import peakbag
from .ellone import ellone
from .star import star
from .mcmc import mcmc
